'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
from enum import Enum
from typing import Optional, Dict, Any

import pandas as pd

from pm4py.util import exec_utils, constants, xes_constants, pandas_utils
from pm4py.util import typing
from pm4py.utils import is_polars_lazyframe


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ZETA = "zeta"
    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def apply(
    df: pd.DataFrame,
    temporal_profile: typing.TemporalProfile,
    parameters: Optional[Dict[Any, Any]] = None,
) -> typing.TemporalProfileConformanceResults:
    """
    Checks the conformance of the dataframe using the provided temporal profile.

    Implements the approach described in:
    Stertz, Florian, Jürgen Mangler, and Stefanie Rinderle-Ma. "Temporal Conformance Checking at Runtime based on Time-infused Process Models." arXiv preprint arXiv:2008.07262 (2020).


    Parameters
    ---------------
    df
        Pandas dataframe
    temporal_profile
        Temporal profile
    parameters
        Parameters of the algorithm, including:
         - Parameters.ACTIVITY_KEY => the attribute to use as activity
         - Parameters.START_TIMESTAMP_KEY => the attribute to use as start timestamp
         - Parameters.TIMESTAMP_KEY => the attribute to use as timestamp
         - Parameters.ZETA => multiplier for the standard deviation
         - Parameters.CASE_ID_KEY => column to use as case identifier

    Returns
    ---------------
    list_dev
        A list containing, for each case, all the deviations.
        Each deviation is a tuple with four elements:
        - 1) The source activity of the recorded deviation
        - 2) The target activity of the recorded deviation
        - 3) The time passed between the occurrence of the source activity and the target activity
        - 4) The value of (time passed - mean)/std for this occurrence (zeta).

    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    zeta = exec_utils.get_param_value(Parameters.ZETA, parameters, 6.0)

    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )
    workcalendar = exec_utils.get_param_value(
        Parameters.WORKCALENDAR,
        parameters,
        constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    )

    temporal_profile = pandas_utils.instantiate_dataframe(
        [
            {
                activity_key: x[0],
                activity_key + "_2": x[1],
                "@@min": y[0] - zeta * y[1],
                "@@max": y[0] + zeta * y[1],
                "@@mean": y[0],
                "@@std": y[1],
            }
            for x, y in temporal_profile.items()
        ]
    )

    if is_polars_lazyframe(df):
        import polars as pl  # type: ignore[import-untyped]

        cases = (
            df.select(pl.col(case_id_key).unique())
            .collect()
            .get_column(case_id_key)
            .to_list()
        )
    else:
        cases = pandas_utils.format_unique(df[case_id_key].unique())

    ret = [[] for c in cases]

    if is_polars_lazyframe(df):
        from pm4py.algo.discovery.dfg.adapters.polars.df_statistics import get_partial_order_dataframe
    else:
        from pm4py.algo.discovery.dfg.adapters.pandas.df_statistics import get_partial_order_dataframe

    efg = get_partial_order_dataframe(
        df,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        start_timestamp_key=start_timestamp_key,
        case_id_glue=case_id_key,
        keep_first_following=False,
        business_hours=business_hours,
        business_hours_slot=business_hours_slots,
        workcalendar=workcalendar,
    )

    if is_polars_lazyframe(df):
        flow_time_col = constants.DEFAULT_FLOW_TIME
        temporal_profile_pl = pl.from_pandas(temporal_profile).lazy()

        efg = (
            efg.select(
                [
                    pl.col(case_id_key),
                    pl.col(activity_key),
                    pl.col(activity_key + "_2"),
                    pl.col(flow_time_col),
                ]
            )
            .join(
                temporal_profile_pl,
                on=[activity_key, activity_key + "_2"],
                how="inner",
            )
            .filter(
                (pl.col(flow_time_col) < pl.col("@@min"))
                | (pl.col(flow_time_col) > pl.col("@@max"))
            )
            .select(
                [
                    pl.col(case_id_key),
                    pl.col(activity_key),
                    pl.col(activity_key + "_2"),
                    pl.col(flow_time_col).alias("@@flow_time"),
                    pl.col("@@mean"),
                    pl.col("@@std"),
                ]
            )
            .collect()
            .to_dicts()
        )
    else:
        efg = efg[[case_id_key, activity_key, activity_key + "_2", "@@flow_time"]]
        efg = efg.merge(temporal_profile, on=[activity_key, activity_key + "_2"])
        efg = efg[
            (efg["@@flow_time"] < efg["@@min"])
            | (efg["@@flow_time"] > efg["@@max"])
        ][
            [
                case_id_key,
                activity_key,
                activity_key + "_2",
                "@@flow_time",
                "@@mean",
                "@@std",
            ]
        ].to_dict(
            "records"
        )

    for el in efg:
        this_zeta = (
            abs(el["@@flow_time"] - el["@@mean"]) / el["@@std"]
            if el["@@std"] > 0
            else sys.maxsize
        )
        ret[cases.index(el[case_id_key])].append(
            (
                el[activity_key],
                el[activity_key + "_2"],
                el["@@flow_time"],
                this_zeta,
            )
        )

    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.objects.conversion.log import converter as log_conversion
from pm4py.objects.log.exporter.xes.variants import etree_xes_exp, line_by_line
from pm4py.util import exec_utils


class Variants(Enum):
    ETREE = etree_xes_exp
    LINE_BY_LINE = line_by_line


DEFAULT_VARIANT = Variants.LINE_BY_LINE


def apply(log, output_file_path, variant=DEFAULT_VARIANT, parameters=None):
    """
    Method to export a XES from a log

    Parameters
    -----------
    log
        Trace log
    output_file_path
        Output file path
    variant
        Selected variant of the algorithm
    parameters
        Parameters of the algorithm:
            Parameters.COMPRESS -> Indicates that the XES file must be compressed
    """
    parameters = dict() if parameters is None else parameters
    return exec_utils.get_variant(variant).apply(
        log_conversion.apply(
            log,
            variant=log_conversion.Variants.TO_EVENT_LOG,
            parameters=parameters,
        ),
        output_file_path,
        parameters=parameters,
    )


def serialize(log, variant=DEFAULT_VARIANT, parameters=None):
    """
    Serialize a log into a binary string containing the XES of the log

    Parameters
    -----------
    log
        Trace log
    variant
        Selected variant of the algorithm
    parameters
        Parameters of the algorithm

    Returns
    -----------
    string
        String describing the XES
    """
    parameters = dict() if parameters is None else parameters

    log_string = exec_utils.get_variant(variant).export_log_as_string(
        log_conversion.apply(
            log,
            variant=log_conversion.Variants.TO_EVENT_LOG,
            parameters=parameters,
        ),
        parameters=parameters,
    )

    return log_string



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
class Interval:
    def __init__(self, start, end, data=None):
        if start >= end:
            raise ValueError("Start must be less than end")
        self.start = start
        self.end = end
        self.data = data or {}

    def overlaps(self, qstart, qend):
        return self.start < qend and self.end > qstart

    def contains(self, point):
        return self.start <= point <= self.end

    def __repr__(self):
        return f"Interval({self.start}, {self.end}, data={self.data})"

class IntervalTree:
    class Node:
        def __init__(self, interval):
            self.interval = interval
            self.max_end = interval.end
            self.left = None
            self.right = None

    def __init__(self):
        self.root = None

    def add(self, interval):
        if self.root is None:
            self.root = self.Node(interval)
            return

        stack = []
        curr = self.root
        while curr:
            stack.append(curr)
            if interval.start < curr.interval.start:
                if curr.left is None:
                    curr.left = self.Node(interval)
                    break
                curr = curr.left
            else:
                if curr.right is None:
                    curr.right = self.Node(interval)
                    break
                curr = curr.right

        # Update max_end from the parent up to the root
        while stack:
            curr = stack.pop()
            curr.max_end = curr.interval.end
            if curr.left:
                curr.max_end = max(curr.max_end, curr.left.max_end)
            if curr.right:
                curr.max_end = max(curr.max_end, curr.right.max_end)

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step is not None:
                raise ValueError("Step in slicing is not supported")
            qstart, qend = key.start, key.stop
            if qstart >= qend:
                return []
            return self._query_interval(self.root, qstart, qend)
        elif isinstance(key, (int, float)):
            return self._query_point(self.root, key)
        else:
            raise ValueError("Only slicing or point queries are supported")

    def _query_interval(self, node, qstart, qend):
        if node is None:
            return []

        result = []
        stack = [node]
        while stack:
            current = stack.pop()
            if current.interval.overlaps(qstart, qend):
                result.append(current.interval)
            if current.right and current.interval.start < qend:
                stack.append(current.right)
            if current.left and current.left.max_end >= qstart:
                stack.append(current.left)
        return result

    def _query_point(self, node, point):
        if node is None:
            return []

        result = []
        stack = [node]
        while stack:
            current = stack.pop()
            if current.interval.contains(point):
                result.append(current.interval)
            if current.right and current.interval.start <= point:
                stack.append(current.right)
            if current.left and current.left.max_end >= point:
                stack.append(current.left)
        return result



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.algo.clustering.profiles.variants import sklearn_profiles
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from typing import Optional, Dict, Any, Generator, Union


class Variants(Enum):
    SKLEARN_PROFILES = sklearn_profiles


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    variant=Variants.SKLEARN_PROFILES,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Generator[EventLog, None, None]:
    """
    Apply clustering to the provided event log
    (methods based on the extraction of profiles for the traces of the event log)

    Implements the approach described in:
    Song, Minseok, Christian W. Günther, and Wil MP Van der Aalst. "Trace clustering in process mining." Business Process Management Workshops: BPM 2008 International Workshops, Milano, Italy, September 1-4, 2008. Revised Papers 6. Springer Berlin Heidelberg, 2009.

    Parameters
    ----------------
    log
        Event log
    variant
        Variant of the clustering to be used, available values:
        - Variants.SKLEARN_PROFILES
    parameters
        Variant-specific parameters

    Returns
    ----------------
    generator
        Generator of dataframes (clusters)
    """
    return exec_utils.get_variant(variant).apply(log, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
def merge_groups_based_on_activities(a, b, groups):
    group_a = None
    group_b = None
    for group in groups:
        if a in group:
            group_a = group
        if b in group:
            group_b = group
    groups = [
        group for group in groups if group != group_a and group != group_b
    ]
    groups.append(group_a.union(group_b))
    return groups


def merge_lists_based_on_activities(a, b, groups):
    group_a = []
    group_b = []
    for group in groups:
        if a in group:
            group_a = group
        if b in group:
            group_b = group
    if group_a is group_b:
        return groups
    groups = [
        group for group in groups if group != group_a and group != group_b
    ]
    groups.append(group_a + group_b)
    return groups



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional


def connect(email_user: Optional[str], mailbox_id: int):
    """
    Returns a mailbox object from the local Outlook instance.

    Parameters
    -------------
    email_user
        E-mail address to use
    mailbox_id
        ID of the mailbox to use:
        * 5 = outbox
        * 6 = inbox
        * 9 = calendar

    Returns
    -------------
    mailbox_obj
        Mailbox object
    """
    import pythoncom

    pythoncom.CoInitialize()

    import win32com.client

    outlook = win32com.client.Dispatch("Outlook.Application").GetNamespace(
        "MAPI"
    )
    mailbox_id = int(mailbox_id)

    if email_user is not None:
        recipient = outlook.CreateRecipient(email_user)
        recipient.Resolve()
        return outlook.GetSharedDefaultFolder(recipient, mailbox_id)

    return outlook.GetDefaultFolder(mailbox_id)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import defaultdict

class PrefixSpan:
    def __init__(self, projection):
        self.projection = projection

    def frequent(self, min_occ):
        results = self._prefix_span([], self.projection, min_occ)
        results.sort(key=lambda x: x[1], reverse=True)
        inverted = [(supp, pattern) for pattern, supp in results]
        return inverted

    def _prefix_span(self, prefix, proj_db, min_occ):
        results = []
        freq_items = self._find_freq_items(proj_db, min_occ)
        for item, supp in sorted(freq_items.items()):  # Sort items for consistent order
            new_prefix = prefix + [item]
            results.append((new_prefix, supp))
            new_proj_db = self._project(proj_db, item)
            results.extend(self._prefix_span(new_prefix, new_proj_db, min_occ))
        return results

    def _find_freq_items(self, proj_db, min_occ):
        item_count = defaultdict(int)
        for seq in proj_db:
            seen = set()
            for item in seq:
                if item not in seen:
                    item_count[item] += 1
                    seen.add(item)
        return {item: count for item, count in item_count.items() if count >= min_occ}

    def _project(self, proj_db, item):
        new_db = []
        for seq in proj_db:
            found = False
            for i in range(len(seq)):
                if not found and seq[i] == item:
                    found = True
                    suffix = seq[i+1:]
                    if suffix:
                        new_db.append(suffix)
                    break
        return new_db



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import shutil
from enum import Enum
from typing import Optional, Dict, Any

from pm4py.util import exec_utils, vis_utils, constants
from pm4py.visualization.performance_spectrum.variants import neato


class Variants(Enum):
    NEATO = neato


def apply(
    perf_spectrum: Dict[str, Any],
    variant=Variants.NEATO,
    parameters: Optional[Dict[Any, Any]] = None,
) -> str:
    """
    Construct the performance spectrum visualization

    Parameters
    ----------------
    perf_spectrum
        Performance spectrum
    variant
        Variant of the visualization to use:
        - NEATO: using the Graphviz Neato layouter
    parameters
        Variant-specific parameters

    Returns
    ---------------
    file_path
        Path containing the visualization
    """
    return exec_utils.get_variant(variant).apply(
        perf_spectrum, parameters=parameters
    )


def view(figure: str):
    """
    Views the performance spectrum

    Parameters
    ---------------
    figure
        Path containing the visualization
    """
    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        if constants.DEFAULT_GVIZ_VIEW == "matplotlib_view":
            import matplotlib.pyplot as plt
            import matplotlib.image as mpimg

            img = mpimg.imread(figure)
            plt.axis("off")
            plt.tight_layout(pad=0, w_pad=0, h_pad=0)
            plt.imshow(img)
            plt.show()
            return

        if vis_utils.check_visualization_inside_jupyter():
            vis_utils.view_image_in_jupyter(figure)
        else:
            vis_utils.open_opsystem_image_viewer(figure)


def save(figure: str, output_file_path: str):
    """
    Saves the performance spectrum at the specified path

    Parameters
    ---------------
    figure
        Path containing the visualization
    output_file_path
        Path into which the image should be saved
    """
    shutil.copyfile(figure, output_file_path)
    return ""


def serialize(figure: str):
    """
    Serializes the performance spectrum visualization

    Parameters
    ---------------
    figure
        Path containing the visualization
    """
    with open(figure, "rb") as f:
        return f.read()


def matplotlib_view(figure: str):
    """
    Views the performance spectrum using Matplotlib

    Parameters
    ---------------
    figure
        Path containing the visualization
    """
    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg

        img = mpimg.imread(figure)
        plt.imshow(img)
        plt.show()



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
import importlib.util
from typing import List, Union


def levenshtein_distance(s1, s2):
    if len(s1) < len(s2):
        return levenshtein_distance(s2, s1)

    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def levenshtein(stru1, stru2):
    if importlib.util.find_spec("stringdist"):
        import stringdist

        return stringdist.levenshtein(stru1, stru2)

    return levenshtein_distance(stru1, stru2)


def argmin_levenshtein(stru: str, list_stri: List[str]) -> Union[str, None]:
    """
    Given a string (stru), finds a string in a list
    of strings (list_stri) that minimizes the Levenshtein distance.

    Parameters
    --------------
    stru
         String (that is compared)
    list_stri
        List of comparison strings

    Returns
    --------------
    argmin_dist
        String (belonging to list_stri) that minimizes the Levenshtein distance
        with the 'stru' argument
    """
    if list_stri:
        len_stru = len(stru)
        argmin_dist = None
        min_edit_dist = sys.maxsize
        # sort the strings in the list based on the actual length in comparison
        # to the provided string
        list_stri = sorted(list_stri, key=lambda x: abs(len(x) - len_stru))
        for comp_stri in list_stri:
            this_length_diff = abs(len(comp_stri) - len_stru)
            # to make things faster, breaks whether we reach a string
            # which distance is greater or equal than the edit distance
            # with a previous string, because the edit distance would then
            # be trivially greater or equal in length
            if this_length_diff >= min_edit_dist:
                break
            dist_this_comp = levenshtein(stru, comp_stri)
            if dist_this_comp < min_edit_dist:
                argmin_dist = comp_stri
                min_edit_dist = dist_this_comp
        return argmin_dist
    return None


def argmax_levenshtein(stru: str, list_stri: List[str]) -> Union[str, None]:
    """
    Given a string (stru), finds a string in a list
    of strings (list_stri) that maximizes the Levenshtein distance.

    Parameters
    --------------
    stru
         String (that is compared)
    list_stri
        List of comparison strings

    Returns
    --------------
    argmax_dist
        String (belonging to list_stri) that maximizes the Levenshtein distance
        with the 'stru' argument
    """
    if list_stri:
        len_stru = len(stru)
        argmax_dist = None
        max_edit_dist = 0
        # sort the strings in the list based on the actual length in comparison
        # to the provided string
        list_stri = sorted(list_stri, key=lambda x: -len(x))
        for comp_stri in list_stri:
            # to make things faster, breaks whether the
            # following condition is reached
            if len(comp_stri) + len_stru <= max_edit_dist:
                break
            dist_this_comp = levenshtein(stru, comp_stri)
            if dist_this_comp > max_edit_dist:
                argmax_dist = comp_stri
                max_edit_dist = dist_this_comp
        return argmax_dist
    return None



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys

from pm4py.objects.random_variables.uniform.random_variable import Uniform


class Constant0(Uniform):
    """
    Describes a constant0-equal-to-0 random variable
    """

    def __init__(self):
        """
        Constructor

        Parameters
        ----------
        value
            Constant value of the distribution
        """
        Uniform.__init__(self, loc=0, scale=0)
        self.priority = 1

    def read_from_string(self, distribution_parameters):
        """
        Initialize distribution parameters from string

        Parameters
        -----------
        distribution_parameters
            Current distribution parameters as exported on the Petri net
        """
        return None

    def get_transition_type(self):
        """
        Get the type of transition associated to the current distribution

        Returns
        -----------
        transition_type
            String representing the type of the transition
        """
        return "IMMEDIATE"

    def get_distribution_type(self):
        """
        Get current distribution type

        Returns
        -----------
        distribution_type
            String representing the distribution type
        """
        return "IMMEDIATE"

    def get_distribution_parameters(self):
        """
        Get a string representing distribution parameters

        Returns
        -----------
        distribution_parameters
            String representing distribution parameters
        """
        return "UNDEFINED"

    def get_value(self):
        """
        Get a random value following the distribution

        Returns
        -----------
        value
            Value obtained following the distribution
        """
        return 0

    def get_values(self, no_values=400):
        """
        Get some random values following the distribution

        Parameters
        -----------
        no_values
            Number of values to return

        Returns
        ----------
        values
            Values extracted according to the probability distribution
        """
        return [self.get_value() for i in range(no_values)]

    def calculate_loglikelihood(self, values, tol=0.0001):
        """
        Calculate log likelihood

        Parameters
        ------------
        values
            Empirical values to work on
        tol
            Tolerance about float values (consider them 0?)

        Returns
        ------------
        likelihood
            Log likelihood that the values follows the distribution
        """
        values_0 = [x for x in values if abs(x) < tol]
        if len(values) == len(values_0):
            return sys.float_info.max
        return -sys.float_info.max



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict
from typing import Tuple, List, Any

import numpy as np

from pm4py.objects.log.obj import Trace
from pm4py.objects.petri_net.utils import (
    align_utils,
    petri_utils as petri_utils,
)
from pm4py.objects.petri_net.utils.consumption_matrix import ConsumptionMatrix
from pm4py.objects.petri_net.utils.incidence_matrix import IncidenceMatrix
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.util import exec_utils, constants, xes_constants, points_subset
from pm4py.util.lp import solver


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    MAX_K_VALUE = "max_k_value"
    COSTS = "costs"
    SPLIT_IDX = "split_idx"
    INCIDENCE_MATRIX = "incidence_matrix"
    A = "A_matrix"
    CONSUMPTION_MATRIX = "consumption_matrix"
    C = "C_matrix"
    FULL_BOOTSTRAP_REQUIRED = "full_bootstrap_required"


"""
Implements the extended marking equation as explained in the following paper:

van Dongen, Boudewijn F. "Efficiently computing alignments." International Conference on Business Process Management.
Springer, Cham, 2018.
https://link.springer.com/chapter/10.1007/978-3-319-98648-7_12
"""


class ExtendedMarkingEquationSolver(object):
    def __init__(
        self,
        trace: Trace,
        sync_net: PetriNet,
        sync_im: Marking,
        sync_fm: Marking,
        parameters: Optional[Dict[Any, Any]] = None,
    ):
        """
        Constructor

        Parameters
        ---------------
        trace
            Trace
        sync_net
            Synchronous product net
        sync_im
            Initial marking
        sync_fm
            Final marking
        parameters
            Parameters of the algorithm, including:
            - Parameters.CASE_ID_KEY => attribute to use as case identifier
            - Parameters.ACTIVITY_KEY => attribute to use as activity
            - Parameters.COSTS => (if provided) the cost function (otherwise the default cost function is applied)
            - Parameters.SPLIT_IDX => (if provided) the split points as indices of elements of the trace
                (e.g. for ["A", "B", "C", "D", "E"], specifying [1,3] as split points means splitting at "B" and "D").
                If not provided, some split points at uniform distances are found.
            - Parameters.MAX_K_VALUE => the maximum number of split points that is allowed (trim the specified indexes
                if necessary).
            - Parameters.INCIDENCE_MATRIX => (if provided) the incidence matrix associated to the sync product net
            - Parameters.A => (if provided) the A numpy matrix of the incidence matrix
            - Parameters.CONSUMPTION_MATRIX => (if provided) the consumption matrix associated to the sync product net
            - Parameters.C => (if provided) the C numpy matrix of the consumption matrix
            - Parameters.FULL_BOOTSTRAP_REQUIRED => The preset/postset of places/transitions need to be inserted
        """
        if parameters is None:
            parameters = {}

        activity_key = exec_utils.get_param_value(
            Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
        )
        max_k_value = exec_utils.get_param_value(
            Parameters.MAX_K_VALUE, parameters, 5
        )
        costs = exec_utils.get_param_value(Parameters.COSTS, parameters, None)
        split_idx = exec_utils.get_param_value(
            Parameters.SPLIT_IDX, parameters, None
        )
        self.full_bootstrap_required = exec_utils.get_param_value(
            Parameters.FULL_BOOTSTRAP_REQUIRED, parameters, True
        )

        self.trace = [x[activity_key] for x in trace]
        if costs is None:
            costs = align_utils.construct_standard_cost_function(
                sync_net, align_utils.SKIP
            )
        if split_idx is None:
            split_idx = [i for i in range(1, len(trace))]
        self.split_idx = split_idx
        if len(self.split_idx) > max_k_value:
            self.split_idx = points_subset.pick_chosen_points_list(
                max_k_value, self.split_idx
            )
        self.k = len(self.split_idx) if len(self.split_idx) > 1 else 2
        self.sync_net = sync_net
        self.ini = sync_im
        self.fin = sync_fm
        self.costs = costs
        self.incidence_matrix = exec_utils.get_param_value(
            Parameters.INCIDENCE_MATRIX,
            parameters,
            IncidenceMatrix(self.sync_net),
        )
        self.consumption_matrix = exec_utils.get_param_value(
            Parameters.CONSUMPTION_MATRIX,
            parameters,
            ConsumptionMatrix(self.sync_net),
        )
        self.A = exec_utils.get_param_value(
            Parameters.A,
            parameters,
            np.asmatrix(self.incidence_matrix.a_matrix),
        )
        self.C = exec_utils.get_param_value(
            Parameters.C,
            parameters,
            np.asmatrix(self.consumption_matrix.c_matrix),
        )

        self.__build_entities()

    def __build_entities(self):
        """
        Builds entities useful to define the marking equation
        """
        self.Aeq = None
        self.__build_encodings()
        if self.full_bootstrap_required:
            petri_utils.decorate_transitions_prepostset(self.sync_net)
            petri_utils.decorate_places_preset_trans(self.sync_net)

    def __build_encodings(self):
        """
        Encodes the aforementioned objects in Numpy matrixes
        """
        self.zeros = np.zeros((self.A.shape[0], self.A.shape[1]))
        self.ini_vec = np.matrix(
            self.incidence_matrix.encode_marking(self.ini)
        ).transpose()
        self.fin_vec = np.matrix(
            self.incidence_matrix.encode_marking(self.fin)
        ).transpose()
        transitions = self.incidence_matrix.transitions
        self.inv_indices = {y: x for x, y in transitions.items()}
        self.inv_indices = [
            self.inv_indices[i] for i in range(len(self.inv_indices))
        ]
        self.c = self.__build_cost_vector()
        self.__build_variable_corr()
        self.__build_h_value_cost_vector()
        self.c1 = list(self.c)
        self.__build_non_null_entries_y()

    def __build_cost_vector(self) -> List[int]:
        """
        Builds the complete cost vector of the integer problem
        """
        c1 = [
            self.costs[self.inv_indices[i]]
            for i in range(len(self.inv_indices))
        ] * self.k
        c2 = [
            self.costs[self.inv_indices[i]]
            for i in range(len(self.inv_indices))
        ] * (self.k - 1)
        c3 = [0 for i in range(len(self.inv_indices))] * (self.k - 1)
        c = c1 + c2 + c3
        return c

    def __build_h_value_cost_vector(self):
        """
        Builds the cost vector for the heuristics calculation as explained in the paper
        """
        self.c0 = list(self.c)
        for idx in self.y[-1]:
            self.c0[idx] = 0

    def __build_variable_corr(self):
        """
        The variables of the LP are split between both "x", "y" and "xy" (which is the sum of x and y)

        Make sure we can reconstruct that!
        """
        count = 0
        self.x = []
        for i in range(self.k):
            self.x.append([])
            for j in range(len(self.inv_indices)):
                self.x[-1].append(count)
                count = count + 1
        self.y = []
        for i in range(self.k - 1):
            self.y.append([])
            for j in range(len(self.inv_indices)):
                self.y[-1].append(count)
                count = count + 1
        self.xy = []
        for i in range(self.k - 1):
            self.xy.append([])
            for j in range(len(self.inv_indices)):
                self.xy[-1].append(count)
                count = count + 1

    def __build_Aeq1(self) -> np.ndarray:
        """
        Builds point (1) of the extended marking equation paper
        """
        Aeq1 = np.hstack(
            [self.A]
            + (2 * self.k - 2) * [self.zeros]
            + (self.k - 1) * [self.A]
        )
        return Aeq1

    def __build_Aeq2(self) -> np.ndarray:
        """
        We have some variables xy, that are the sum of x and y.
        Just make sure that xy = x + y
        (not implementing any specific point!)
        """
        Aeq_stack = []
        for i in range(1, len(self.x)):
            Aeq = np.zeros((len(self.x[i]), self.xy[-1][-1] + 1))
            for j in range(len(self.x[i])):
                Aeq[j][self.x[i][j]] = 1
                Aeq[j][self.y[i - 1][j]] = 1
                Aeq[j][self.xy[i - 1][j]] = -1
            Aeq_stack.append(Aeq)
        Aeq = np.vstack(Aeq_stack)
        return Aeq

    def __build_non_null_entries_y(self):
        """
        Utility function that is needed for point (6) of the extended marking equation paper
        """
        self.null_entries = []
        for i in self.split_idx:
            self.null_entries.append([])
            act = self.trace[i]
            for j in range(len(self.inv_indices)):
                if self.inv_indices[j].label[0] == act:
                    self.null_entries[-1].append(j)

    def __build_Aeq3(self) -> np.ndarray:
        """
        Implements point (6) of the extended marking equation paper
        """
        Aeq_stack = []
        for i in range(len(self.y)):
            Aeq = np.zeros((1, self.xy[-1][-1] + 1))
            for j in range(len(self.y[i])):
                Aeq[0][self.y[i][j]] = 1
            Aeq_stack.append(Aeq)
        Aeq = np.vstack(Aeq_stack)
        return Aeq

    def __build_Aeq4(self) -> np.ndarray:
        """
        Implements point (5) of the extended marking equation paper
        """
        Aeq = np.zeros((len(self.null_entries), self.xy[-1][-1] + 1))
        for i in range(min(len(self.null_entries), len(self.y))):
            for j in range(len(self.y[i])):
                if j not in self.null_entries[i]:
                    Aeq[i][self.y[i][j]] = 1
        return Aeq

    def __build_Aub1(self) -> np.ndarray:
        """
        Implements points (3) and (4) of the extended marking equation paper
        """
        Aub_stack = []
        for i in range(len(self.x)):
            Aub = np.zeros((len(self.x[i]), self.xy[-1][-1] + 1))
            for j in range(len(self.x[i])):
                Aub[j][self.x[i][j]] = -1
            Aub_stack.append(Aub)
        for i in range(len(self.y)):
            Aub = np.zeros((len(self.y[i]), self.xy[-1][-1] + 1))
            for j in range(len(self.y[i])):
                Aub[j][self.y[i][j]] = -1
            Aub_stack.append(Aub)
        Aub = np.vstack(Aub_stack)
        return Aub

    def __build_Aub2(self) -> np.ndarray:
        """
        Implements point (2) of the extended marking equation paper
        """
        Aub_stack = []
        for i in range(self.k - 1):
            v = [self.A]
            v = v + (self.k - 1) * [self.zeros]
            for j in range(self.k - 1):
                if j == i:
                    v = v + [self.C]
                else:
                    v = v + [self.zeros]
            for j in range(self.k - 1):
                if j < i:
                    v = v + [self.A]
                else:
                    v = v + [self.zeros]
            Aub = np.hstack(v)
            Aub_stack.append(Aub)
        Aub = np.vstack(Aub_stack)
        Aub = -Aub
        return Aub

    def __build_lin_prog_components(self):
        """
        Builds the components needed to solve the marking equation
        """
        Aeq1 = self.__build_Aeq1()
        Aeq2 = self.__build_Aeq2()
        Aeq3 = self.__build_Aeq3()
        Aeq4 = self.__build_Aeq4()
        self.Aeq = np.vstack([Aeq1, Aeq2, Aeq3, Aeq4])
        Aub1 = self.__build_Aub1()
        Aub2 = self.__build_Aub2()
        self.Aub = np.vstack([Aub1, Aub2])

        if (
            solver.DEFAULT_LP_SOLVER_VARIANT
            == solver.CVXOPT_SOLVER_CUSTOM_ALIGN
        ):
            from cvxopt import matrix

            self.Aeq_transf = matrix(self.Aeq.astype(np.float64))
            self.Aub_transf = matrix(self.Aub.astype(np.float64))
            self.c = matrix([1.0 * x for x in self.c])
        else:
            self.Aeq_transf = self.Aeq
            self.Aub_transf = self.Aub

    def __calculate_vectors(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates the inequality/equality vector from the knowledge of the
        initial/final marking of the synchronous product net
        """
        beq_stack = []
        bub_stack = []
        beq_stack.append(self.fin_vec - self.ini_vec)
        beq_stack.append(np.zeros(((len(self.x) - 1) * len(self.x[0]), 1)))
        beq_stack.append(np.ones((len(self.y), 1)))
        beq_stack.append(np.zeros((len(self.null_entries), 1)))
        bub_stack.append(np.zeros((len(self.x) * len(self.x[0]), 1)))
        bub_stack.append(np.zeros((len(self.y) * len(self.y[0]), 1)))
        for i in range(self.k - 1):
            bub_stack.append(self.ini_vec)

        beq = np.vstack(beq_stack)
        bub = np.vstack(bub_stack)

        return beq, bub

    def get_components(self) -> Tuple[Any, Any, Any, Any, Any]:
        """
        Retrieve the components (Numpy matrixes) of the problem

        Returns
        ---------------
        c
            objective function
        Aub
           Inequalities matrix
        bub
            Inequalities vector
        Aeq
            Equalities matrix
        beq
            Equalities vector
        """
        if self.Aeq is None:
            self.__build_lin_prog_components()

        self.beq, self.bub = self.__calculate_vectors()

        if (
            solver.DEFAULT_LP_SOLVER_VARIANT
            == solver.CVXOPT_SOLVER_CUSTOM_ALIGN
        ):
            from cvxopt import matrix

            self.beq_transf = matrix(self.beq.astype(np.float64))
            self.bub_transf = matrix(self.bub.astype(np.float64))
        else:
            self.beq_transf = self.beq
            self.bub_transf = self.bub

        return (
            self.c,
            self.Aub_transf,
            self.bub_transf,
            self.Aeq_transf,
            self.beq_transf,
        )

    def change_ini_vec(self, ini: Marking):
        """
        Changes the initial marking of the synchronous product net

        Parameters
        --------------
        ini
            Initial marking
        """
        self.ini = ini
        self.ini_vec = np.matrix(
            self.incidence_matrix.encode_marking(ini)
        ).transpose()

    def get_x_vector(self, sol_points: List[int]) -> List[int]:
        """
        Returns the x vector of the solution

        Parameters
        --------------
        sol_points
            Solution of the integer problem

        Returns
        ---------------
        x
            X vector
        """
        x_vector = [0] * len(self.sync_net.transitions)
        for i in range(len(self.x)):
            for j in range(len(self.x[i])):
                x_vector[j] += sol_points[self.x[i][j]]
        for i in range(len(self.y)):
            for j in range(len(self.y[i])):
                x_vector[j] += sol_points[self.y[i][j]]
        return x_vector

    def get_h(self, sol_points: List[int]) -> int:
        """
        Returns the value of the heuristics

        Parameters
        --------------
        sol_points
            Solution of the integer problem

        Returns
        --------------
        h
            Heuristics value
        """
        h = 0.0
        for i in range(len(sol_points)):
            h += sol_points[i] * self.c1[i]
        return int(h)

    def get_activated_transitions(
        self, sol_points: List[int]
    ) -> List[PetriNet.Transition]:
        """
        Gets the transitions of the synchronous product net that are non-zero
        in the solution of the marking equation

        Parameters
        --------------
        sol_points
            Solution of the integer problem

        Returns
        --------------
        act_trans
            Activated transitions
        """
        act_trans = []
        for i in range(len(sol_points)):
            for j in range(sol_points[i]):
                act_trans.append(self.inv_indices[i])
        return act_trans

    def solve(self, variant=None) -> Tuple[int, List[int]]:
        """
        Solves the extended marking equation, returning the heuristics and the x vector

        Parameters
        -------------
        variant
            Variant of the ILP solver to use

        Returns
        -------------
        h
            Heuristics value
        x
            X vector
        """
        if variant is None:
            variant = solver.DEFAULT_LP_SOLVER_VARIANT
        # use ILP solver
        if variant is solver.CVXOPT_SOLVER_CUSTOM_ALIGN:
            variant = solver.CVXOPT_SOLVER_CUSTOM_ALIGN_ILP
        c, Aub, bub, Aeq, beq = self.get_components()
        parameters_solver = {}
        parameters_solver["use_ilp"] = True
        sol = solver.apply(
            c,
            Aub,
            bub,
            Aeq,
            beq,
            variant=variant,
            parameters=parameters_solver,
        )
        sol_points = solver.get_points_from_sol(sol, variant=variant)
        if sol_points is not None:
            x = self.get_x_vector(sol_points)
            x = [int(y) for y in x]
            h = self.get_h(sol_points)
            return h, x
        return None, None

    def get_firing_sequence(
        self, x: List[int]
    ) -> Tuple[List[PetriNet.Transition], bool, int]:
        """
        Gets a firing sequence from the X vector

        Parameters
        ----------------
        x
            X vector

        Returns
        ----------------
        firing_sequence
            Firing sequence
        reach_fm
            Boolean value that is true whether the firing sequence reaches the final marking
        explained_events
            Number of explaned events by the firing sequence
        """
        activated_transitions = self.get_activated_transitions(x)
        firing_sequence, reach_fm, explained_events = (
            align_utils.search_path_among_sol(
                self.sync_net, self.ini, self.fin, activated_transitions
            )
        )
        return firing_sequence, reach_fm, explained_events


def build(
    trace: Trace,
    sync_net: PetriNet,
    sync_im: Marking,
    sync_fm: Marking,
    parameters: Optional[Dict[Any, Any]] = None,
) -> ExtendedMarkingEquationSolver:
    """
    Builds the extended marking equation out of a trace and a synchronous product net

    Parameters
    ---------------
    trace
        Trace
    sync_net
        Synchronous product net
    sync_im
        Initial marking (of sync net)
    sync_fm
        Final marking (of sync net)
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY => attribute to use as case identifier
        - Parameters.ACTIVITY_KEY => attribute to use as activity
        - Parameters.COSTS => (if provided) the cost function (otherwise the default cost function is applied)
        - Parameters.SPLIT_IDX => (if provided) the split points as indices of elements of the trace
            (e.g. for ["A", "B", "C", "D", "E"], specifying [1,3] as split points means splitting at "B" and "D").
            If not provided, some split points at uniform distances are found.
        - Parameters.MAX_K_VALUE => the maximum number of split points that is allowed (trim the specified indexes
            if necessary).
        - Parameters.INCIDENCE_MATRIX => (if provided) the incidence matrix associated to the sync product net
        - Parameters.A => (if provided) the A numpy matrix of the incidence matrix
        - Parameters.CONSUMPTION_MATRIX => (if provided) the consumption matrix associated to the sync product net
        - Parameters.C => (if provided) the C numpy matrix of the consumption matrix
        - Parameters.FULL_BOOTSTRAP_REQUIRED => The preset/postset of places/transitions need to be inserted
    """
    if parameters is None:
        parameters = {}

    return ExtendedMarkingEquationSolver(
        trace, sync_net, sync_im, sync_fm, parameters=parameters
    )


def get_h_value(
    solver: ExtendedMarkingEquationSolver,
    parameters: Optional[Dict[Any, Any]] = None,
) -> int:
    """
    Gets the heuristics value from the extended marking equation

    Parameters
    --------------
    solver
        Extended marking equation solver (class in this file)
    parameters
        Possible parameters of the algorithm
    """
    return solver.solve()[0]



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import TypeVar, Generic, Dict, Any, Optional

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureUVCL,
    IMDataStructureLog,
)
from pm4py.algo.discovery.inductive.fall_through.empty_traces import (
    EmptyTracesUVCL,
)
from pm4py.algo.discovery.inductive.variants.abc import InductiveMinerFramework
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.dfg.obj import DFG
from copy import copy
from enum import Enum
from pm4py.util import exec_utils


T = TypeVar("T", bound=IMDataStructureLog)


class IMFParameters(Enum):
    NOISE_THRESHOLD = "noise_threshold"


class IMF(Generic[T], InductiveMinerFramework[T]):

    def instance(self) -> IMInstance:
        return IMInstance.IMf


class IMFUVCL(IMF[IMDataStructureUVCL]):
    def apply(
        self,
        obj: IMDataStructureUVCL,
        parameters: Optional[Dict[str, Any]] = None,
        second_iteration: bool = False,
    ) -> ProcessTree:
        noise_threshold = exec_utils.get_param_value(
            IMFParameters.NOISE_THRESHOLD, parameters, 0.0
        )

        empty_traces = EmptyTracesUVCL.apply(obj, parameters)
        if empty_traces is not None and empty_traces[1]:
            number_original_traces = sum(
                y for y in obj.data_structure.values()
            )
            number_filtered_traces = sum(
                y for y in empty_traces[1][1].data_structure.values()
            )

            if (
                number_original_traces - number_filtered_traces
                > noise_threshold * number_original_traces
            ):
                return self._recurse(
                    empty_traces[0], empty_traces[1], parameters
                )
            else:
                obj = empty_traces[1][1]

        tree = self.apply_base_cases(obj, parameters)
        if tree is None:
            cut = self.find_cut(obj, parameters)
            if cut is not None:
                tree = self._recurse(cut[0], cut[1], parameters=parameters)
            if tree is None:
                if not second_iteration:
                    filtered_ds = self.__filter_dfg_noise(obj, noise_threshold)
                    tree = self.apply(
                        filtered_ds,
                        parameters=parameters,
                        second_iteration=True,
                    )
                    if tree is None:
                        ft = self.fall_through(obj, parameters)
                        tree = self._recurse(
                            ft[0], ft[1], parameters=parameters
                        )
        return tree

    def __filter_dfg_noise(self, obj, noise_threshold):
        start_activities = copy(obj.dfg.start_activities)
        end_activities = copy(obj.dfg.end_activities)
        dfg = copy(obj.dfg.graph)
        outgoing_max_occ = {}
        for x, y in dfg.items():
            act = x[0]
            if act not in outgoing_max_occ:
                outgoing_max_occ[act] = y
            else:
                outgoing_max_occ[act] = max(y, outgoing_max_occ[act])
            if act in end_activities:
                outgoing_max_occ[act] = max(
                    outgoing_max_occ[act], end_activities[act]
                )
        dfg_list = sorted(
            [(x, y) for x, y in dfg.items()],
            key=lambda x: (x[1], x[0]),
            reverse=True,
        )
        dfg_list = [
            x
            for x in dfg_list
            if x[1] > noise_threshold * outgoing_max_occ[x[0][0]]
        ]
        dfg_list = [x[0] for x in dfg_list]
        # filter the elements in the DFG
        graph = {x: y for x, y in dfg.items() if x in dfg_list}

        dfg = DFG()
        for sa in start_activities:
            dfg.start_activities[sa] = start_activities[sa]
        for ea in end_activities:
            dfg.end_activities[ea] = end_activities[ea]
        for act in graph:
            dfg.graph[act] = graph[act]

        return IMDataStructureUVCL(obj.data_structure, dfg)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, Union, List
import numpy as np
from pm4py.util.regex import SharedObj, get_new_char
from pm4py.util import string_distance
from pm4py.util import exec_utils
from scipy.optimize import linprog
import importlib.util


class Parameters:
    STRING_DISTANCE = "string_distance"
    USE_FAST_EMD = "use_fast_emd"  # New parameter to choose between implementations


def normalized_levensthein(s1, s2):
    return float(string_distance.levenshtein(s1, s2)) / float(
        max(len(s1), len(s2))
    )


def get_act_correspondence(activities, parameters=None):
    if parameters is None:
        parameters = {}

    shared_obj = SharedObj()
    ret = {}
    for act in activities:
        get_new_char(act, shared_obj)
        ret[act] = shared_obj.mapping_dictio[act]

    return ret


def encode_two_languages(lang1, lang2, parameters=None):
    if parameters is None:
        parameters = {}

    all_activities = sorted(
        list(
            set(y for x in lang1 for y in x).union(
                set(y for x in lang2 for y in x)
            )
        )
    )
    acts_corresp = get_act_correspondence(
        all_activities, parameters=parameters
    )

    enc1 = {}
    enc2 = {}

    for k in lang1:
        new_key = "".join(acts_corresp[i] for i in k)
        enc1[new_key] = lang1[k]

    for k in lang2:
        new_key = "".join(acts_corresp[i] for i in k)
        enc2[new_key] = lang2[k]

    for x in enc1:
        if x not in enc2:
            enc2[x] = 0.0

    for x in enc2:
        if x not in enc1:
            enc1[x] = 0.0

    enc1 = [(x, y) for x, y in enc1.items()]
    enc2 = [(x, y) for x, y in enc2.items()]

    enc1 = sorted(enc1, reverse=True, key=lambda x: x[0])
    enc2 = sorted(enc2, reverse=True, key=lambda x: x[0])

    return enc1, enc2


class EMDCalculator:
    """
    A class that provides an EMD (Earth Mover's Distance) computation similar to what `pyemd` offers.
    It uses linear programming via `scipy.optimize.linprog` to solve the underlying flow problem.

    Usage:
    ------
    emd_value = EMDCalculator.emd(first_histogram, second_histogram, distance_matrix)
    """

    @staticmethod
    def emd(
            first_histogram: np.ndarray,
            second_histogram: np.ndarray,
            distance_matrix: np.ndarray,
    ) -> float:
        """
        Compute the Earth Mover's Distance given two histograms and a distance matrix.

        Parameters
        ----------
        first_histogram : np.ndarray
            The first distribution (array of nonnegative numbers).
        second_histogram : np.ndarray
            The second distribution (array of nonnegative numbers).
        distance_matrix : np.ndarray
            Matrix of distances between points of the two distributions.

        Returns
        -------
        float
            The computed EMD value.
        """
        # Ensure the histograms sum to the same total
        sum1 = np.sum(first_histogram)
        sum2 = np.sum(second_histogram)
        if not np.isclose(sum1, sum2):
            raise ValueError(
                "Histograms must sum to the same total for EMD calculation."
            )

        n = len(first_histogram)
        m = len(second_histogram)

        # Flatten the distance matrix
        c = distance_matrix.flatten()

        # Constraints:
        # sum_j F_ij = first_histogram[i] for each i
        # sum_i F_ij = second_histogram[j] for each j
        # F_ij >= 0

        # We have n "row sum" constraints and m "column sum" constraints.
        A_eq = []
        b_eq = []

        # Row constraints
        for i in range(n):
            row_constraint = np.zeros(n * m)
            for j in range(m):
                row_constraint[i * m + j] = 1
            A_eq.append(row_constraint)
            b_eq.append(first_histogram[i])

        # Column constraints
        for j in range(m):
            col_constraint = np.zeros(n * m)
            for i in range(n):
                col_constraint[i * m + j] = 1
            A_eq.append(col_constraint)
            b_eq.append(second_histogram[j])

        A_eq = np.array(A_eq)
        b_eq = np.array(b_eq)

        # Bounds: F_ij >= 0
        bounds = [(0, None) for _ in range(n * m)]

        # Solve the LP:
        # minimize c^T x subject to A_eq x = b_eq and x >= 0
        res = linprog(c, A_eq=A_eq, b_eq=b_eq, bounds=bounds, method="highs")

        if res.status != 0:
            raise ValueError(
                "Linear programming failed. Status: {}, Message: {}".format(
                    res.status, res.message))

        # The optimal value is the EMD
        return res.fun


class POTEMDCalculator:
    """
    A faster implementation of EMD using the POT (Python Optimal Transport) library.
    Falls back to the SciPy implementation if POT is not available.

    Usage:
    ------
    emd_value = POTEMDCalculator.emd(first_histogram, second_histogram, distance_matrix)
    """

    @staticmethod
    def is_pot_available():
        """Check if POT (Python Optimal Transport) is available."""
        return importlib.util.find_spec("ot") is not None

    @staticmethod
    def emd(
            first_histogram: np.ndarray,
            second_histogram: np.ndarray,
            distance_matrix: np.ndarray,
    ) -> float:
        """
        Compute the Earth Mover's Distance using POT if available, otherwise fall back to SciPy.

        Parameters
        ----------
        first_histogram : np.ndarray
            The first distribution (array of nonnegative numbers).
        second_histogram : np.ndarray
            The second distribution (array of nonnegative numbers).
        distance_matrix : np.ndarray
            Matrix of distances between points of the two distributions.

        Returns
        -------
        float
            The computed EMD value.
        """
        # Check if POT is available
        if POTEMDCalculator.is_pot_available():
            try:
                import ot

                # Normalize histograms to sum to 1.0 if they don't sum to the same value
                sum1 = np.sum(first_histogram)
                sum2 = np.sum(second_histogram)

                if not np.isclose(sum1, sum2):
                    # Normalize both histograms to sum to 1.0
                    a = first_histogram / sum1
                    b = second_histogram / sum2
                else:
                    a = first_histogram
                    b = second_histogram

                # Ensure arrays have the right type
                a = np.asarray(a, dtype=np.float64)
                b = np.asarray(b, dtype=np.float64)
                M = np.asarray(distance_matrix, dtype=np.float64)

                # Use POT's EMD computation
                # Regularized version can be used with reg=1e-3 for efficiency when needed
                emd_value = ot.emd2(a, b, M)

                return emd_value

            except ImportError:
                # Fall back to SciPy implementation
                return EMDCalculator.emd(first_histogram, second_histogram, distance_matrix)
            except Exception as e:
                # If POT fails for any reason, fall back to SciPy implementation
                print(f"POT EMD calculation failed: {str(e)}. Falling back to SciPy implementation.")
                return EMDCalculator.emd(first_histogram, second_histogram, distance_matrix)
        else:
            # Fall back to SciPy implementation
            return EMDCalculator.emd(first_histogram, second_histogram, distance_matrix)


def apply(
        lang1: Dict[List[str], float],
        lang2: Dict[List[str], float],
        parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    if parameters is None:
        parameters = {}

    distance_function = exec_utils.get_param_value(
        Parameters.STRING_DISTANCE, parameters, normalized_levensthein
    )

    # New parameter to choose between implementations
    use_fast_emd = exec_utils.get_param_value(
        Parameters.USE_FAST_EMD, parameters, True
    )

    enc1, enc2 = encode_two_languages(lang1, lang2, parameters=parameters)

    first_histogram = np.array([x[1] for x in enc1])
    second_histogram = np.array([x[1] for x in enc2])

    distance_matrix = []
    for x in enc1:
        row = []
        for y in enc2:
            dist = distance_function(x[0], y[0])
            row.append(float(dist))
        distance_matrix.append(row)

    distance_matrix = np.array(distance_matrix)

    # Choose which EMD implementation to use
    if use_fast_emd and POTEMDCalculator.is_pot_available():
        ret = POTEMDCalculator.emd(first_histogram, second_histogram, distance_matrix)
    else:
        # Use the original SciPy implementation
        ret = EMDCalculator.emd(first_histogram, second_histogram, distance_matrix)

    return ret



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import defaultdict
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Tuple, Set


def apply(ocel: OCEL,
          parameters: Optional[Dict[Any,
                                    Any]] = None) -> Tuple[Set[str],
                                                           Dict[Tuple[str,
                                                                str,
                                                                str],
                                                                int]]:
    if parameters is None:
        parameters = {}

    import pm4py

    # Available graph types
    graph_types = [
        "object_interaction",
        "object_descendants",
        "object_inheritance",
        "object_cobirth",
        "object_codeath"
    ]

    # Initialize OTG components
    edges = defaultdict(int)

    object_types = set(ocel.objects["ocel:type"].unique())
    objects = ocel.objects.to_dict("records")
    objects = {x["ocel:oid"]: x["ocel:type"] for x in objects}

    # Iterate over each relationship type
    for sigma in graph_types:
        # Discover the object graph for the relationship type
        object_graph = pm4py.discover_objects_graph(ocel, graph_type=sigma)

        # Build the edges for the OTG
        for o1, o2 in object_graph:
            ot1 = objects[o1]
            ot2 = objects[o2]

            ot_pair = (ot1, ot2)

            edge = (ot_pair[0], sigma, ot_pair[1])
            edges[edge] += 1

    return object_types, dict(edges)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Union, Optional, Dict, Any

import pandas as pd

from pm4py.algo.discovery.minimum_self_distance.variants import log, pandas
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.util import exec_utils, pandas_utils
from pm4py.utils import is_polars_lazyframe
import importlib.util

POLARS_AVAILABLE = importlib.util.find_spec("polars") is not None
if POLARS_AVAILABLE:
    from pm4py.algo.discovery.minimum_self_distance.variants import polars


class Variants(Enum):
    LOG = log
    PANDAS = pandas
    POLARS = polars if POLARS_AVAILABLE else log


def apply(
    log_obj: Union[EventLog, pd.DataFrame, EventStream],
    variant: Union[str, None] = None,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, int]:
    if parameters is None:
        parameters = {}

    if variant is None:
        if is_polars_lazyframe(log_obj):
            if not POLARS_AVAILABLE:
                raise RuntimeError(
                    "Polars LazyFrame provided but 'polars' package is not installed."
                )
            variant = Variants.POLARS
        elif pandas_utils.check_is_pandas_dataframe(log_obj):
            variant = Variants.PANDAS
        else:
            variant = Variants.LOG

    return exec_utils.get_variant(variant).apply(
        log_obj, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from graphviz import Digraph
from enum import Enum
from pm4py.util import exec_utils
from pm4py.visualization.ocel.ocdfg.variants import classic, elkjs
from typing import Optional, Dict, Any


class Variants(Enum):
    CLASSIC = classic
    ELKJS = elkjs


def apply(
    ocdfg: Dict[str, Any],
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Digraph:
    """
    Visualizes an OC-DFG using one of the provided visualizations.

    Parameters
    ----------------
    ocdfg
        Object-centric directly-follows graph
    variant
        Available variants. Possible values:
        - Variants.CLASSIC
    parameters
        Variant-specific parameters

    Returns
    ----------------
    viz
        Graphviz DiGraph
    """
    return exec_utils.get_variant(variant).apply(ocdfg, parameters)


def save(
    gviz, output_file_path: str, variant=Variants.CLASSIC, parameters=None
):
    """
    Saves the diagram
    """
    return exec_utils.get_variant(variant).save(
        gviz, output_file_path, parameters
    )


def view(gviz, variant=Variants.CLASSIC, parameters=None):
    """
    Views the diagram
    """
    return exec_utils.get_variant(variant).view(gviz, parameters)


def matplotlib_view(gviz, variant=Variants.CLASSIC, parameters=None):
    """
    Views the diagram using Matplotlib
    """
    return exec_utils.get_variant(variant).matplotlib_view(gviz, parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import datetime
from copy import copy
from enum import Enum
from typing import Optional, Dict, Any, Union

from pm4py.objects import petri_net
from pm4py.objects.log import obj as log_instance
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils import (
    final_marking as final_marking_discovery,
)
from pm4py.util import constants
from pm4py.util import exec_utils
from pm4py.util import xes_constants
from pm4py.util.dt_parsing.variants import strpfromiso


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    RETURN_VISITED_ELEMENTS = "return_visited_elements"
    NO_TRACES = "noTraces"
    MAX_TRACE_LENGTH = "maxTraceLength"
    LOG = "log"
    STOCHASTIC_MAP = "smap"
    PETRI_SEMANTICS = "petri_semantics"


def apply_playout(
    net,
    initial_marking,
    no_traces=100,
    max_trace_length=100,
    case_id_key=xes_constants.DEFAULT_TRACEID_KEY,
    activity_key=xes_constants.DEFAULT_NAME_KEY,
    timestamp_key=xes_constants.DEFAULT_TIMESTAMP_KEY,
    final_marking=None,
    smap=None,
    log=None,
    return_visited_elements=False,
    semantics=petri_net.semantics.ClassicSemantics(),
    parameters=None,
):
    """
    Do the playout of a Petrinet generating a log

    Parameters
    ----------
    net
        Petri net to play-out
    initial_marking
        Initial marking of the Petri net
    no_traces
        Number of traces to generate
    max_trace_length
        Maximum number of events per trace (do break)
    case_id_key
        Trace attribute that is the case ID
    activity_key
        Event attribute that corresponds to the activity
    timestamp_key
        Event attribute that corresponds to the timestamp
    final_marking
        If provided, the final marking of the Petri net
    smap
        Stochastic map
    log
        Log
    semantics
        Semantics of the Petri net to be used (default: petri_net.semantics.ClassicSemantics())
    """
    from pm4py.algo.simulation.montecarlo.utils import replay
    from pm4py.objects.stochastic_petri import utils as stochastic_utils

    if final_marking is None:
        # infer the final marking from the net
        final_marking = final_marking_discovery.discover_final_marking(net)
    if smap is None:
        if log is None:
            raise Exception(
                "please provide at least one between stochastic map and log"
            )
        if parameters is None:
            parameters = {}
        parameters_rep = copy(parameters)
        parameters_rep[Parameters.ACTIVITY_KEY] = activity_key
        parameters_rep[Parameters.TIMESTAMP_KEY] = timestamp_key
        smap = replay.get_map_from_log_and_net(
            log, net, initial_marking, final_marking, parameters=parameters_rep
        )

    # assigns to each event an increased timestamp from 1970
    curr_timestamp = 10000000
    all_visited_elements = []

    for i in range(no_traces):
        visited_elements = []
        visible_transitions_visited = []
        marking = copy(initial_marking)

        while len(visible_transitions_visited) < max_trace_length:
            visited_elements.append(marking)

            if not semantics.enabled_transitions(
                net, marking
            ):  # supports nets with possible deadlocks
                break
            all_enabled_trans = semantics.enabled_transitions(net, marking)
            if final_marking is not None and marking == final_marking:
                en_t_list = list(all_enabled_trans.union({None}))
            else:
                en_t_list = list(all_enabled_trans)

            trans = stochastic_utils.pick_transition(en_t_list, smap)

            if trans is None:
                break

            visited_elements.append(trans)
            if trans.label is not None:
                visible_transitions_visited.append(trans)

            marking = semantics.execute(trans, net, marking)

        all_visited_elements.append(tuple(visited_elements))

    if return_visited_elements:
        return all_visited_elements

    log = log_instance.EventLog()

    for index, visited_elements in enumerate(all_visited_elements):
        trace = log_instance.Trace()
        trace.attributes[case_id_key] = str(index)
        for element in visited_elements:
            if (
                type(element) is PetriNet.Transition
                and element.label is not None
            ):
                event = log_instance.Event()
                event[activity_key] = element.label
                event[timestamp_key] = strpfromiso.fix_naivety(
                    datetime.datetime.fromtimestamp(curr_timestamp)
                )
                trace.append(event)
                # increases by 1 second
                curr_timestamp += 1
        log.append(trace)

    return log


def apply(
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking = None,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Do the playout of a Petrinet generating a log

    Parameters
    -----------
    net
        Petri net to play-out
    initial_marking
        Initial marking of the Petri net
    final_marking
        If provided, the final marking of the Petri net
    parameters
        Parameters of the algorithm:
            Parameters.NO_TRACES -> Number of traces of the log to generate
            Parameters.MAX_TRACE_LENGTH -> Maximum trace length
            Parameters.PETRI_SEMANTICS -> Petri net semantics to be used (default: petri_nets.semantics.ClassicSemantics())
    """
    if parameters is None:
        parameters = {}
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, xes_constants.DEFAULT_TRACEID_KEY
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    no_traces = exec_utils.get_param_value(
        Parameters.NO_TRACES, parameters, 1000
    )
    max_trace_length = exec_utils.get_param_value(
        Parameters.MAX_TRACE_LENGTH, parameters, 1000
    )
    smap = exec_utils.get_param_value(
        Parameters.STOCHASTIC_MAP, parameters, None
    )
    log = exec_utils.get_param_value(Parameters.LOG, parameters, None)
    return_visited_elements = exec_utils.get_param_value(
        Parameters.RETURN_VISITED_ELEMENTS, parameters, False
    )
    semantics = exec_utils.get_param_value(
        Parameters.PETRI_SEMANTICS,
        parameters,
        petri_net.semantics.ClassicSemantics(),
    )

    return apply_playout(
        net,
        initial_marking,
        max_trace_length=max_trace_length,
        no_traces=no_traces,
        case_id_key=case_id_key,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        final_marking=final_marking,
        smap=smap,
        log=log,
        return_visited_elements=return_visited_elements,
        semantics=semantics,
        parameters=None,
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Any, Dict, List, Tuple, Union, Optional, Collection
import networkx as nx
import numpy as np
from pm4py.objects.process_tree.obj import ProcessTree, Operator
from pm4py.objects.process_tree.utils.generic import is_leaf, is_tau_leaf
from pm4py.util.lp import solver
from pm4py.utils import project_on_event_attribute
import pandas as pd
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util import constants, xes_constants, thread_utils
from pm4py.objects.log.obj import EventLog
import importlib.util
from functools import lru_cache
from collections import defaultdict


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    SHOW_PROGRESS_BAR = "show_progress_bar"


class ProcessTreeAligner:
    """
    Implementation of the approach described in:

    Schwanen, Christopher T., Wied Pakusa, and Wil MP van der Aalst. "Process tree alignments." Enterprise Design, Operations, and Computing, ser. LNCS, Cham: Springer International Publishing (2024).
    """

    def __init__(self, tree: ProcessTree):
        self.tree = tree
        self.graph = nx.MultiDiGraph()
        self.node_id_counter = 0  # Initialize node ID counter
        # Cache to store activity-to-edges mapping
        self.activity_to_edges = defaultdict(list)
        # Flag for tracking if the tree has been processed
        self._tree_processed = False
        self._build_process_tree_graph(self.tree)
        self._precompute_activity_edges()

    def _precompute_activity_edges(self):
        """Precompute activity to edges mapping for faster lookups during alignment"""
        for edge in self.graph.edges(keys=True, data=True):
            u, v, k, data = edge
            label = data.get('label')
            if label is not None:
                self.activity_to_edges[label].append((u, v, k))

    def _get_new_node_id(self) -> int:
        node_id = self.node_id_counter
        self.node_id_counter += 1
        return node_id

    def _build_process_tree_graph(self, tree: ProcessTree) -> None:
        start_node = self._get_new_node_id()
        self.graph.add_node(start_node, source=True)

        if is_tau_leaf(tree):
            self.graph.nodes[start_node]["sink"] = True
            return

        if tree.operator == Operator.LOOP:
            if len(tree.children) != 2:
                raise Exception(f"Loop {tree} does not have exactly two children")
            if is_tau_leaf(tree.children[0]):
                # Special handling when the first child of loop is a tau transition
                self.graph.nodes[start_node]["sink"] = True
                self._build_process_tree_subgraph(tree.children[1], start_node, start_node)
            else:
                end_node = self._get_new_node_id()
                self.graph.add_node(end_node, sink=True)
                self._build_process_tree_subgraph(tree.children[0], start_node, end_node)
                self._build_process_tree_subgraph(tree.children[1], end_node, start_node)
        else:
            end_node = self._get_new_node_id()
            self.graph.add_node(end_node, sink=True)
            self._build_process_tree_subgraph(tree, start_node, end_node)

        self._tree_processed = True

    def _build_process_tree_subgraph(self, tree: ProcessTree, start_node: Any, end_node: Any, iac: int = 1) -> None:
        if tree.operator is None:
            self._build_process_tree_subgraph_leaf(tree, start_node, end_node, iac)
        elif tree.operator == Operator.SEQUENCE:
            self._build_process_tree_subgraph_sequence(tree, start_node, end_node, iac)
        elif tree.operator == Operator.XOR:
            self._build_process_tree_subgraph_xor(tree, start_node, end_node, iac)
        elif tree.operator == Operator.PARALLEL:
            self._build_process_tree_subgraph_parallel(tree, start_node, end_node, iac)
        elif tree.operator == Operator.LOOP:
            self._build_process_tree_subgraph_loop(tree, start_node, end_node, iac)
        else:
            raise Exception(f"Operator {tree.operator} is not supported")

    def _build_process_tree_subgraph_leaf(self, tree: ProcessTree, start_node: Any, end_node: Any, iac: int) -> None:
        if not is_leaf(tree):
            raise Exception(f"Subtree {tree} is not a leaf")
        self.graph.add_edge(
            start_node,
            end_node,
            label=tree.label,
            capacity=1. / iac,
            cost=iac if tree.label is not None else 0
        )

    def _build_process_tree_subgraph_sequence(self, tree: ProcessTree, start_node: Any, end_node: Any,
                                              iac: int) -> None:
        if len(tree.children) == 1:
            self._build_process_tree_subgraph(tree.children[0], start_node, end_node, iac)
            return

        nodes = [start_node] + [self._get_new_node_id() for _ in range(len(tree.children) - 1)] + [end_node]
        for i in range(len(tree.children)):
            self._build_process_tree_subgraph(tree.children[i], nodes[i], nodes[i + 1], iac)

    def _build_process_tree_subgraph_xor(self, tree: ProcessTree, start_node: Any, end_node: Any, iac: int) -> None:
        for child in tree.children:
            self._build_process_tree_subgraph(child, start_node, end_node, iac)

    def _build_process_tree_subgraph_parallel(self, tree: ProcessTree, start_node: Any, end_node: Any,
                                              iac: int) -> None:
        if tree.operator != Operator.PARALLEL:
            raise Exception(f"Operator {tree.operator} is not a parallel")

        # Ensure start_node and end_node are added to the graph
        if start_node not in self.graph.nodes:
            self.graph.add_node(start_node)
        if end_node not in self.graph.nodes:
            self.graph.add_node(end_node)

        if 'shuffle' not in self.graph.nodes[start_node]:
            self.graph.nodes[start_node]["shuffle"] = []
            self.graph.nodes[start_node]["iac"] = iac
            self.graph.nodes[start_node]["is_split"] = True
        if 'shuffle' not in self.graph.nodes[end_node]:
            self.graph.nodes[end_node]["shuffle"] = []
            self.graph.nodes[end_node]["iac"] = iac
            self.graph.nodes[end_node]["is_join"] = True

        shuffle_split = []
        shuffle_join = []
        local_iac = iac * len(tree.children)

        for child in tree.children:
            spread_node_start = self._get_new_node_id()
            spread_node_end = self._get_new_node_id()
            self.graph.add_node(spread_node_start)
            self.graph.add_node(spread_node_end)

            # Record shuffle edges
            shuffle_split.append((start_node, spread_node_start, 0))
            shuffle_join.append((spread_node_end, end_node, 0))

            # Add shuffle edges with appropriate capacity
            self.graph.add_edge(
                start_node, spread_node_start,
                label=None, capacity=1. / local_iac, cost=0, shuffle=True
            )
            self.graph.add_edge(
                spread_node_end, end_node,
                label=None, capacity=1. / local_iac, cost=0, shuffle=True
            )

            # Build subgraph for the child
            self._build_process_tree_subgraph(child, spread_node_start, spread_node_end, local_iac)

        self.graph.nodes[start_node]["shuffle"].append(shuffle_split)
        self.graph.nodes[end_node]["shuffle"].append(shuffle_join)

    def _build_process_tree_subgraph_loop(self, tree: ProcessTree, start_node: Any, end_node: Any, iac: int) -> None:
        old_start_node = start_node
        start_node = self._get_new_node_id()
        self.graph.add_node(start_node)
        self.graph.add_edge(
            old_start_node, start_node,
            label=None, capacity=1. / iac, cost=0
        )
        old_end_node = end_node
        end_node = self._get_new_node_id()
        self.graph.add_node(end_node)
        self.graph.add_edge(
            end_node, old_end_node,
            label=None, capacity=1. / iac, cost=0
        )

        self._build_process_tree_subgraph(tree.children[0], start_node, end_node, iac)
        self._build_process_tree_subgraph(tree.children[1], end_node, start_node, iac)

    @lru_cache(maxsize=128)
    def align(self, trace: Tuple[str]) -> Tuple[float, List[Tuple[str, str]]]:
        """
        Align a trace with the process tree. Using tuple for trace to enable caching.

        Args:
            trace: A tuple of activity labels (converted from list for caching)

        Returns:
            A tuple containing (alignment cost, alignment moves)
        """
        # Convert back to list for processing
        trace_list = list(trace)
        return self._align_impl(trace_list)

    def _align_impl(self, trace: List[str]) -> Tuple[float, List[Tuple[str, str]]]:
        """Implementation of the alignment algorithm"""
        align_variant = "pulp"

        num_steps = len(trace) + 1
        graph = self.graph

        # Pre-processed data
        edges = list(graph.edges(keys=True))
        nodes = list(graph.nodes())

        # Use vectorized operations and efficient data structures
        # Variable indexing
        var_index = {}
        var_counter = 0

        # Using dictionaries instead of individual variables for better memory management
        x_vars = {}  # Flow variables
        y_vars = {}  # Log move variables
        z_vars = {}  # Sync move variables
        s_vars = {}  # Shuffle variables

        # Pre-allocate arrays for improved performance
        # Approximate size of arrays based on problem dimensions
        max_vars = num_steps * (len(edges) + len(nodes) + len(edges) + len(nodes))
        c = np.zeros(max_vars)  # Objective function coefficients

        # x variables
        for i in range(num_steps):
            for e in edges:
                x_vars[(i,) + e] = var_counter
                var_index[var_counter] = ('x', i, e)
                var_counter += 1

        # y variables (continuous between 0 and 1)
        for i in range(1, num_steps):
            for v in nodes:
                y_vars[(i, v)] = var_counter
                var_index[var_counter] = ('y', i, v)
                var_counter += 1

        # z variables (continuous between 0 and capacity)
        sync_edges = {}
        for idx, activity in enumerate(trace):
            step = idx + 1
            # Use precomputed activity to edges mapping for better performance
            matching_edges = self.activity_to_edges.get(activity, [])
            sync_edges[step] = matching_edges
            for e in matching_edges:
                z_vars[(step,) + e] = var_counter
                var_index[var_counter] = ('z', step, e)
                var_counter += 1

        # s variables (binary)
        shuffles = {}
        for v in nodes:
            node_data = graph.nodes[v]
            if node_data.get('shuffle'):
                for idx, shuffle_edges in enumerate(node_data['shuffle']):
                    shuffles[(v, idx)] = shuffle_edges
        for i in range(num_steps):
            for key in shuffles.keys():
                s_vars[(i, key)] = var_counter
                var_index[var_counter] = ('s', i, key)
                var_counter += 1

        num_vars = var_counter

        # Resize c to actual number of variables
        c = np.zeros(num_vars)

        # For x variables - vectorized assignment when possible
        for key, idx in x_vars.items():
            i, u, v, k = key
            cost = graph.edges[(u, v, k)].get('cost', 0)
            c[idx] = cost

        # For y variables - all have cost 1
        for idx in y_vars.values():
            c[idx] = 1

        # For z variables
        for key, idx in z_vars.items():
            step, u, v, k = key
            edge_cost = graph.edges[(u, v, k)].get('cost', 0)
            c[idx] = 1 - edge_cost if edge_cost > 1 else 0

        # Constraints
        # Use more efficient sparse matrix construction
        Aeq_rows = []
        beq = []

        # Flow conservation constraints - vectorize when possible
        for i in range(num_steps):
            for v in nodes:
                row = {}
                rhs = 0

                # Sum of inflow arcs
                for e in edges:
                    if e[1] == v:
                        idx = x_vars.get((i,) + e)
                        if idx is not None:
                            row[idx] = row.get(idx, 0) + 1

                # Sum of outflow arcs
                for e in edges:
                    if e[0] == v:
                        idx = x_vars.get((i,) + e)
                        if idx is not None:
                            row[idx] = row.get(idx, 0) - 1

                # Add y and z variables
                if i > 0:
                    idx_y = y_vars.get((i, v))
                    if idx_y is not None:
                        row[idx_y] = row.get(idx_y, 0) + 1
                    for e in edges:
                        if e[1] == v:
                            idx_z = z_vars.get((i,) + e)
                            if idx_z is not None:
                                row[idx_z] = row.get(idx_z, 0) + 1
                if i < num_steps - 1:
                    idx_y = y_vars.get((i + 1, v))
                    if idx_y is not None:
                        row[idx_y] = row.get(idx_y, 0) - 1
                    for e in edges:
                        if e[0] == v:
                            idx_z = z_vars.get((i + 1,) + e)
                            if idx_z is not None:
                                row[idx_z] = row.get(idx_z, 0) - 1

                # Handle source and sink nodes
                if i == 0 and graph.nodes[v].get('source'):
                    rhs = -1
                elif i == len(trace) and graph.nodes[v].get('sink'):
                    rhs = 1

                if row:
                    Aeq_rows.append((row, rhs))

        # Shuffle constraints
        for key, shuffle_edges in shuffles.items():
            v, idx = key
            iac = graph.nodes[v]['iac']
            for i in range(num_steps):
                row = {}
                for u, w, _ in shuffle_edges:
                    edge = (u, w, 0)
                    idx_x = x_vars.get((i,) + edge)
                    if idx_x is not None:
                        row[idx_x] = row.get(idx_x, 0) + 1
                idx_s = s_vars.get((i, key))
                if idx_s is not None:
                    row[idx_s] = row.get(idx_s, 0) - (1.0 / iac)
                Aeq_rows.append((row, 0))

        # Duplicate labels constraints
        Aub_rows = []
        bub = []
        for i in sync_edges:
            if len(sync_edges[i]) > 1:
                row = {}
                for e in sync_edges[i]:
                    idx_z = z_vars.get((i,) + e)
                    if idx_z is not None:
                        edge_cost = graph.edges[e].get('cost', 0)
                        row[idx_z] = edge_cost
                Aub_rows.append((row, 1))

        # Variable bounds
        lb = np.zeros(num_vars)
        ub = np.full(num_vars, np.inf)

        # x variables
        for key, idx in x_vars.items():
            i, u, v, k = key
            capacity = graph.edges[(u, v, k)].get('capacity', np.inf)
            ub[idx] = capacity

        # y variables
        for idx in y_vars.values():
            ub[idx] = 1

        # z variables
        for key, idx in z_vars.items():
            i, u, v, k = key
            capacity = graph.edges[(u, v, k)].get('capacity', np.inf)
            ub[idx] = capacity

        # s variables
        for idx in s_vars.values():
            ub[idx] = 1
            lb[idx] = 0

        # Variable types
        vartype = [0] * num_vars
        for idx in s_vars.values():
            vartype[idx] = 1

        # Build Aeq efficiently with COO format for sparse matrix creation
        Aeq_data = []
        Aeq_row_idx = []
        Aeq_col_idx = []
        for row_idx, (row, rhs_value) in enumerate(Aeq_rows):
            for var_idx, coeff in row.items():
                Aeq_data.append(coeff)
                Aeq_row_idx.append(row_idx)
                Aeq_col_idx.append(var_idx)
            beq.append(rhs_value)

        # Create sparse matrix once, then convert to array format after all data is populated
        from scipy import sparse
        Aeq = sparse.csr_matrix((Aeq_data, (Aeq_row_idx, Aeq_col_idx)), shape=(len(Aeq_rows), num_vars))
        Aeq = Aeq.toarray()

        # Build Aub
        Aub_data = []
        Aub_row_idx = []
        Aub_col_idx = []
        for row_idx, (row, rhs_value) in enumerate(Aub_rows):
            for var_idx, coeff in row.items():
                Aub_data.append(coeff)
                Aub_row_idx.append(row_idx)
                Aub_col_idx.append(var_idx)
            bub.append(rhs_value)

        Aub = sparse.csr_matrix((Aub_data, (Aub_row_idx, Aub_col_idx)), shape=(len(Aub_rows), num_vars))
        Aub = Aub.toarray()

        # Solver parameters
        bounds = [(lb[i], ub[i]) for i in range(num_vars)]
        parameters = {
            'bounds': bounds,
            'integrality': vartype,
        }

        c = [float(x) for x in c]
        bub = [float(x) for x in bub]
        beq = [float(x) for x in beq]

        if align_variant == "cvxopt_solver_custom_align_ilp":
            del parameters["bounds"]

            Aub_add = np.zeros((2 * len(bounds), len(c)))

            for idx, b in enumerate(bounds):
                Aub_add[2 * idx, idx] = -1.0
                Aub_add[2 * idx + 1, idx] = 1.0
                bub.append(-b[0])
                bub.append(b[1])

            Aub = np.vstack([Aub, Aub_add])

            bub = np.asarray(bub)[:, np.newaxis]
            beq = np.asarray(beq)[:, np.newaxis]

            from cvxopt import matrix

            Aub = matrix(Aub.astype(np.float64))
            bub = matrix(bub)
            Aeq = matrix(Aeq.astype(np.float64))
            beq = matrix(beq)
            c = matrix(c)

        # Solve the LP problem
        try:
            sol = solver.apply(c, Aub, bub, Aeq, beq, variant=align_variant, parameters=parameters)

            prim_obj = solver.get_prim_obj_from_sol(sol, variant=align_variant)
            var_values = solver.get_points_from_sol(sol, variant=align_variant)

            # Reconstruct the alignment moves more efficiently
            alignment_moves = []
            i = 1  # Start from step 1
            while i <= len(trace):
                activity = trace[i - 1]
                move_recorded = False

                # Check for synchronous moves (z variables)
                for e in sync_edges.get(i, []):
                    idx_z = z_vars.get((i,) + e)
                    if idx_z is not None and var_values[idx_z] > 1e-5:
                        # Synchronous move
                        alignment_moves.append((activity, activity))
                        move_recorded = True
                        break

                if move_recorded:
                    i += 1
                    continue

                # Check for moves on log (y variables)
                for v in nodes:
                    idx_y = y_vars.get((i, v))
                    if idx_y is not None and var_values[idx_y] > 1e-5:
                        alignment_moves.append((activity, '>>'))
                        move_recorded = True
                        break

                if move_recorded:
                    i += 1
                    continue

                # If neither z nor y variables are active, it's a move on model
                # Find active x variables at position i
                model_moves = []
                for e in edges:
                    idx_x = x_vars.get((i,) + e)
                    if idx_x is not None and var_values[idx_x] > 1e-5:
                        label = graph.edges[e].get('label')
                        if label is not None and label not in model_moves:
                            model_moves.append(label)
                if model_moves:
                    for label in model_moves:
                        alignment_moves.append(('>>', label))
                    i += 1  # Advance to the next step
                else:
                    # No move detected; advance to prevent infinite loop
                    alignment_moves.append((activity, '>>'))
                    i += 1

            return prim_obj, alignment_moves
        except Exception as e:
            raise Exception(f"Optimization failed: {str(e)}")


# Helper function to construct progress bar
def _construct_progress_bar(progress_length, parameters):
    if exec_utils.get_param_value(Parameters.SHOW_PROGRESS_BAR, parameters,
                                  constants.SHOW_PROGRESS_BAR) and importlib.util.find_spec("tqdm"):
        if progress_length > 1:
            from tqdm.auto import tqdm
            return tqdm(total=progress_length, desc="aligning log, completed variants :: ")
    return None


# Helper function to destroy progress bar
def _destroy_progress_bar(progress):
    if progress is not None:
        progress.close()
    del progress


def __perform_alignment_computations(v, variants_align, aligner, empty_cost):
    alignment_cost, alignment_moves = aligner.align(v)
    alignment_cost = round(alignment_cost + 10 ** -14, 13)

    fitness = 1.0 - alignment_cost / (empty_cost + len(v)) if (empty_cost + len(v)) > 0 else 0.0

    alignment = {"cost": alignment_cost, "alignment": alignment_moves, "fitness": fitness}
    variants_align[v] = alignment


def apply_list_tuple_activities(list_tuple_activities: List[Collection[str]], aligner: ProcessTreeAligner,
                                parameters: Optional[Dict[Any, Any]] = None) -> List[Dict[str, Any]]:
    """
    Apply the alignment algorithm to a list of activities.
    Optimized to use caching and more efficient data structures.
    """
    if parameters is None:
        parameters = {}

    # Use a set for faster lookup of variants
    variants = set(tuple(v) for v in list_tuple_activities)
    variants_align = {}

    progress = _construct_progress_bar(len(variants), parameters)

    # Pre-compute empty trace alignment
    empty_cost, empty_moves = aligner.align(())
    empty_cost = round(empty_cost + 10 ** -14, 13)

    f = lambda x, y: (__perform_alignment_computations(x, y, aligner, empty_cost), progress.update() if progress is not None else None)
    thm = thread_utils.Pm4pyThreadManager()

    # Process each variant only once
    for v in variants:
        thm.submit(f, v, variants_align)

    thm.join()
    _destroy_progress_bar(progress)

    # Map results back to original list
    return [variants_align[tuple(t)] for t in list_tuple_activities]


def apply(log: Union[pd.DataFrame, EventLog], process_tree: ProcessTree, parameters: Optional[Dict[Any, Any]] = None) -> \
        List[Dict[str, Any]]:
    """
    Aligns an event log against a process tree model, using the approach described in:
    Schwanen, Christopher T., Wied Pakusa, and Wil MP van der Aalst. "Process tree alignments." Enterprise Design, Operations, and Computing, ser. LNCS, Cham: Springer International Publishing (2024).

    Parameters
    ---------------
    log
        Event log or Pandas dataframe
    process_tree
        Process tree
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity
        - Parameters.SHOW_PROGRESS_BAR => shows the progress bar

    Returns
    ---------------
    aligned_traces
        List that contains the alignment for each trace
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)

    list_tuple_activities = project_on_event_attribute(log, activity_key)
    list_tuple_activities = [tuple(x) for x in list_tuple_activities]

    aligner = ProcessTreeAligner(process_tree)

    return apply_list_tuple_activities(list_tuple_activities, aligner, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
def get_new_char(label, shared_obj):
    """
    Get a new single character describing the activity, for the regex

    Parameters
    ------------
    label
        Label of the transition
    shared_obj
        Shared object
    """
    list_to_avoid = [
        "[",
        "]",
        "(",
        ")",
        "*",
        "+",
        "^",
        "?",
        "\r",
        "\n",
        " ",
        "\t",
        "$",
        '"',
        "!",
        "#",
        "&",
        "%",
        "|",
        ".",
        ",",
        ";",
        "-",
        "'",
        "\\",
        "/",
        "{",
        "}",
        "$",
    ]
    shared_obj.count_char = shared_obj.count_char + 1
    while chr(shared_obj.count_char) in list_to_avoid:
        shared_obj.count_char = shared_obj.count_char + 1
    shared_obj.mapping_dictio[label] = chr(shared_obj.count_char)


class SharedObj:
    def __init__(self):
        self.mapping_dictio = None
        if self.mapping_dictio is None:
            self.mapping_dictio = {}
        self.count_char = 0



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from pm4py.util.lp import solver as lp_solver
from pm4py.objects.petri_net.utils import align_utils as utils
from pm4py.objects.petri_net.utils.incidence_matrix import construct
from pm4py.objects.petri_net.utils.petri_utils import (
    decorate_places_preset_trans,
    decorate_transitions_prepostset,
)
from copy import copy
import heapq


def __search(net, ini, fin):
    cost_function = {t: 1 for t in net.transitions}

    decorate_transitions_prepostset(net)
    decorate_places_preset_trans(net)

    incidence_matrix = construct(net)
    ini_vec, fin_vec, cost_vec = utils.__vectorize_initial_final_cost(
        incidence_matrix, ini, fin, cost_function
    )

    closed = set()

    a_matrix = np.asmatrix(incidence_matrix.a_matrix).astype(np.float64)
    g_matrix = -np.eye(len(net.transitions))
    h_cvx = np.matrix(np.zeros(len(net.transitions))).transpose()
    cost_vec = [x * 1.0 for x in cost_vec]

    use_cvxopt = False
    if (
        lp_solver.DEFAULT_LP_SOLVER_VARIANT
        == lp_solver.CVXOPT_SOLVER_CUSTOM_ALIGN
        or lp_solver.DEFAULT_LP_SOLVER_VARIANT
        == lp_solver.CVXOPT_SOLVER_CUSTOM_ALIGN_ILP
    ):
        use_cvxopt = True

    if use_cvxopt:
        # not available in the latest version of PM4Py
        from cvxopt import matrix

        a_matrix = matrix(a_matrix)
        g_matrix = matrix(g_matrix)
        h_cvx = matrix(h_cvx)
        cost_vec = matrix(cost_vec)

    h, x = utils.__compute_exact_heuristic_new_version(
        net,
        a_matrix,
        h_cvx,
        g_matrix,
        cost_vec,
        incidence_matrix,
        ini,
        fin_vec,
        lp_solver.DEFAULT_LP_SOLVER_VARIANT,
        use_cvxopt=use_cvxopt,
    )
    ini_state = utils.SearchTuple(0 + h, 0, h, ini, None, None, x, True)
    open_set = [ini_state]
    heapq.heapify(open_set)
    visited = 0
    queued = 0
    traversed = 0

    trans_empty_preset = set(t for t in net.transitions if len(t.in_arcs) == 0)

    while not len(open_set) == 0:
        curr = heapq.heappop(open_set)

        current_marking = curr.m
        # 11/10/2019 (optimization Y, that was optimization X,
        # but with the good reasons this way): avoid checking markings in the cycle using
        # the __get_alt function, but check them 'on the road'
        already_closed = current_marking in closed
        if already_closed:
            continue

        while not curr.trust:
            h, x = utils.__compute_exact_heuristic_new_version(
                net,
                a_matrix,
                h_cvx,
                g_matrix,
                cost_vec,
                incidence_matrix,
                curr.m,
                fin_vec,
                lp_solver.DEFAULT_LP_SOLVER_VARIANT,
                use_cvxopt=use_cvxopt,
            )

            # 11/10/19: shall not a state for which we compute the exact heuristics be
            # by nature a trusted solution?
            tp = utils.SearchTuple(
                curr.g + h, curr.g, h, curr.m, curr.p, curr.t, x, True
            )
            # 11/10/2019 (optimization ZA) heappushpop is slightly more efficient than pushing
            # and popping separately
            curr = heapq.heappushpop(open_set, tp)
            current_marking = curr.m

        # max allowed heuristics value (27/10/2019, due to the numerical
        # instability of some of our solvers)
        if curr.h > lp_solver.MAX_ALLOWED_HEURISTICS:
            continue

        # 12/10/2019: do it again, since the marking could be changed
        already_closed = current_marking in closed
        if already_closed:
            continue

        # 12/10/2019: the current marking can be equal to the final marking only if the heuristics
        # (underestimation of the remaining cost) is 0. Low-hanging fruits
        if curr.h < 0.01:
            if current_marking == fin:
                return utils.__reconstruct_alignment(
                    curr, visited, queued, traversed
                )

        closed.add(current_marking)
        visited += 1

        possible_enabling_transitions = copy(trans_empty_preset)
        for p in current_marking:
            for t in p.ass_trans:
                possible_enabling_transitions.add(t)

        enabled_trans = [
            t
            for t in possible_enabling_transitions
            if t.sub_marking <= current_marking
        ]

        trans_to_visit_with_cost = [
            (t, cost_function[t]) for t in enabled_trans
        ]

        for t, cost in trans_to_visit_with_cost:
            traversed += 1
            new_marking = utils.add_markings(current_marking, t.add_marking)

            if new_marking in closed:
                continue
            g = curr.g + cost

            queued += 1
            h, x = utils.__derive_heuristic(
                incidence_matrix, cost_vec, curr.x, t, curr.h
            )
            trustable = utils.__trust_solution(x)
            new_f = g + h

            tp = utils.SearchTuple(
                new_f, g, h, new_marking, curr, t, x, trustable
            )
            heapq.heappush(open_set, tp)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import collections

from pm4py.util import exec_utils
from pm4py.objects.log import util as log_util
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.objects.transition_system import (
    obj as ts,
    constants as ts_constants,
)
from enum import Enum
from pm4py.util import constants, pandas_utils
from pm4py.objects.transition_system.obj import TransitionSystem
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.objects.conversion.log import converter as log_conversion
import pandas as pd


class Parameters(Enum):
    VIEW_MULTI_SET = "multiset"
    VIEW_SET = "set"
    VIEW_SEQUENCE = "sequence"

    VIEWS = {VIEW_MULTI_SET, VIEW_SET, VIEW_SEQUENCE}

    DIRECTION_FORWARD = "forward"
    DIRECTION_BACKWARD = "backward"
    DIRECTIONS = {DIRECTION_FORWARD, DIRECTION_BACKWARD}

    PARAM_KEY_VIEW = "view"
    PARAM_KEY_WINDOW = "window"
    PARAM_KEY_DIRECTION = "direction"

    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY

    INCLUDE_DATA = "include_data"


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> TransitionSystem:
    if parameters is None:
        parameters = {}
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    include_data = exec_utils.get_param_value(
        Parameters.INCLUDE_DATA, parameters, False
    )

    if pandas_utils.check_is_pandas_dataframe(log):
        case_id_key = exec_utils.get_param_value(
            Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
        )
        control_flow_log = pandas_utils.get_traces(log, case_id_key, activity_key)
    else:
        log = log_conversion.apply(
            log, parameters, log_conversion.TO_EVENT_LOG
        )
        control_flow_log = log_util.log.project_traces(log, activity_key)

    transition_system = ts.TransitionSystem()
    view_sequence = []
    for i in range(len(control_flow_log)):
        provided_case = log[i] if type(log) is EventLog else None
        view_sequence.append(
            __compute_view_sequence(
                control_flow_log[i], provided_case, parameters=parameters
            )
        )
    for vs in view_sequence:
        __construct_state_path(
            vs, transition_system, include_data=include_data
        )
    return transition_system


def __construct_state_path(
    view_sequence, transition_system, include_data=False
):
    for i in range(0, len(view_sequence) - 1):
        sf = {
            "state": s
            for s in transition_system.states
            if s.name == view_sequence[i][0]
        }
        sf = (
            sf["state"]
            if len(sf) > 0
            else ts.TransitionSystem.State(view_sequence[i][0])
        )
        st = {
            "state": s
            for s in transition_system.states
            if s.name == view_sequence[i + 1][0]
        }
        st = (
            st["state"]
            if len(st) > 0
            else ts.TransitionSystem.State(view_sequence[i + 1][0])
        )
        t = {
            "t": t
            for t in sf.outgoing
            if t.name == view_sequence[i][1]
            and t.from_state == sf
            and t.to_state == st
        }
        if len(t) == 0:
            t = ts.TransitionSystem.Transition(view_sequence[i][1], sf, st)
            sf.outgoing.add(t)
            st.incoming.add(t)
        else:
            t = t["t"]
        if include_data:
            # add the event to the data in both the source state,
            # the sink state and the transition
            sf.data[ts_constants.OUTGOING_EVENTS].append(view_sequence[i][2])
            st.data[ts_constants.INGOING_EVENTS].append(view_sequence[i][2])
            t.data[ts_constants.EVENTS].append(view_sequence[i][2])
        transition_system.states.add(sf)
        transition_system.states.add(st)
        transition_system.transitions.add(t)


def __compute_view_sequence(trace, full_case, parameters):
    view_sequences = list()
    direction = exec_utils.get_param_value(
        Parameters.PARAM_KEY_DIRECTION,
        parameters,
        Parameters.DIRECTION_FORWARD.value,
    )
    window = exec_utils.get_param_value(
        Parameters.PARAM_KEY_WINDOW, parameters, 2
    )
    for i in range(0, len(trace) + 1):
        if direction == Parameters.DIRECTION_FORWARD.value:
            view_sequences.append(
                (
                    __apply_abstr(trace[i: i + window], parameters),
                    trace[i] if i < len(trace) else None,
                    (
                        (full_case, i)
                        if full_case is not None and i < len(full_case)
                        else None
                    ),
                )
            )
        else:
            view_sequences.append(
                (
                    __apply_abstr(trace[max(0, i - window): i], parameters),
                    trace[i] if i < len(trace) else None,
                    (
                        (full_case, i)
                        if full_case is not None and i < len(full_case)
                        else None
                    ),
                )
            )
    return view_sequences


def __apply_abstr(seq, parameters):
    key_view = exec_utils.get_param_value(
        Parameters.PARAM_KEY_VIEW, parameters, Parameters.VIEW_SEQUENCE.value
    )

    if key_view == Parameters.VIEW_SEQUENCE.value:
        return list(seq)
    elif key_view == Parameters.VIEW_MULTI_SET.value:
        return collections.Counter(seq)
    elif key_view == Parameters.VIEW_SET.value:
        return set(seq)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, List, Tuple, Union

import polars as pl

from pm4py.algo.discovery.batches.utils import detection
from pm4py.util import exec_utils, constants, xes_constants, pandas_utils


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    EVENT_ID_KEY = "event_id_key"
    MERGE_DISTANCE = "merge_distance"
    MIN_BATCH_SIZE = "min_batch_size"


def _to_seconds_expr(column_name: str, schema: Dict[str, Any]) -> pl.Expr:
    expr = pl.col(column_name)
    dtype_repr = str(schema[column_name]).lower()
    if "date" in dtype_repr or "time" in dtype_repr:
        return (
            expr.dt.cast_time_unit("ns")
            .dt.timestamp("ns")
            .cast(pl.Float64)
            / 1_000_000_000
        )
    if any(token in dtype_repr for token in ("int", "float")):
        return expr.cast(pl.Float64)
    raise Exception(
        f"Column '{column_name}' must be temporal or numeric to evaluate seconds."
    )


def apply(
    log: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Tuple[Tuple[str, str], int, Dict[str, Any]]]:
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, xes_constants.DEFAULT_RESOURCE_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, timestamp_key
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    event_id_key = exec_utils.get_param_value(
        Parameters.EVENT_ID_KEY, parameters, constants.DEFAULT_INDEX_KEY
    )

    pandas_utils.check_pandas_dataframe_columns(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
        start_timestamp_key=start_timestamp_key,
    )

    schema = log.collect_schema()
    attributes = [
        activity_key,
        resource_key,
        start_timestamp_key,
        timestamp_key,
        case_id_key,
    ]
    log_contains_evidkey = event_id_key in schema
    if log_contains_evidkey:
        attributes.append(event_id_key)

    # ensure attribute order without duplicates
    seen = set()
    ordered_attributes: List[str] = []
    for attr in attributes:
        if attr not in seen:
            seen.add(attr)
            ordered_attributes.append(attr)

    lf = log.select([pl.col(col) for col in ordered_attributes])

    lf = lf.with_columns(
        [
            _to_seconds_expr(start_timestamp_key, schema).alias("__pm4py_start_sec__"),
            _to_seconds_expr(timestamp_key, schema).alias("__pm4py_complete_sec__"),
        ]
    )

    lf = lf.with_row_count("__pm4py_order__")

    aggregation_exprs = [
        pl.col("__pm4py_start_sec__")
        .sort_by("__pm4py_order__")
        .alias("__pm4py_start__"),
        pl.col("__pm4py_complete_sec__")
        .sort_by("__pm4py_order__")
        .alias("__pm4py_complete__"),
        pl.col(case_id_key)
        .sort_by("__pm4py_order__")
        .alias("__pm4py_cases__"),
    ]

    if log_contains_evidkey:
        aggregation_exprs.append(
            pl.col(event_id_key)
            .sort_by("__pm4py_order__")
            .alias("__pm4py_events__")
        )

    grouped = (
        lf.group_by(activity_key, resource_key, maintain_order=True)
        .agg(aggregation_exprs)
        .collect()
    )

    actres_grouping: Dict[Tuple[str, str], List[Tuple[Any, ...]]] = {}

    for row in grouped.iter_rows(named=True):
        key = (row[activity_key], row[resource_key])
        starts = list(row["__pm4py_start__"])
        completes = list(row["__pm4py_complete__"])
        cases = list(row["__pm4py_cases__"])
        events_list = (
            list(row["__pm4py_events__"])
            if log_contains_evidkey
            else None
        )

        batches = []
        for idx in range(len(starts)):
            entry: Tuple[Any, ...]
            if log_contains_evidkey and events_list is not None:
                entry = (starts[idx], completes[idx], cases[idx], events_list[idx])
            else:
                entry = (starts[idx], completes[idx], cases[idx])
            batches.append(entry)

        actres_grouping[key] = batches

    return detection.detect(actres_grouping, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils, constants
from pm4py.objects.ocel.util import filtering_utils
from copy import deepcopy
from typing import Dict, Any, Optional, Collection
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.ocel import constants as ocel_constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE
    TEMP_COLUMN = "temp_column"
    TEMP_SEPARATOR = "temp_separator"


def apply(
    ocel: OCEL,
    correspondence_dict: Dict[str, Collection[str]],
    parameters: Optional[Dict[Any, Any]] = None,
) -> OCEL:
    """
    Filters an object-centric event log keeping only the specified object types
    with the specified activity set (filters out the rest).

    Parameters
    ----------------
    ocel
        Object-centric event log
    correspondence_dict
        Dictionary containing, for every object type of interest, a
        collection of allowed activities.  Example:

        {"order": ["Create Order"], "element": ["Create Order", "Create Delivery"]}

        Keeps only the object types "order" and "element".
        For the "order" object type, only the activity "Create Order" is kept.
        For the "element" object type, only the activities "Create Order" and "Create Delivery" are kept.
    parameters
        Parameters of the algorithm, including:
            - Parameters.ACTIVITY_KEY => the activity key
            - Parameters.OBJECT_TYPE => the object type column

    Returns
    -----------------
    filtered_ocel
        Filtered object-centric event log
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, ocel.event_activity
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, ocel.object_type_column
    )
    temp_column = exec_utils.get_param_value(
        Parameters.TEMP_COLUMN, parameters, "@@temp_column"
    )
    temp_separator = exec_utils.get_param_value(
        Parameters.TEMP_SEPARATOR, parameters, "@#@#"
    )

    ocel = deepcopy(ocel)

    inv_dict = set()
    for ot in correspondence_dict:
        for act in correspondence_dict[ot]:
            inv_dict.add(act + temp_separator + ot)

    ocel.relations[temp_column] = (
        ocel.relations[activity_key]
        + temp_separator
        + ocel.relations[object_type_column]
    )
    ocel.relations = ocel.relations[ocel.relations[temp_column].isin(inv_dict)]

    del ocel.relations[temp_column]

    return filtering_utils.propagate_relations_filtering(
        ocel, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Dict, Any, Optional
from pm4py.objects.ocel.util import ocel_consistency
from pm4py.objects.ocel.util import filtering_utils
import os


def apply(
    ocel: OCEL, target_path: str, parameters: Optional[Dict[Any, Any]] = None
):
    """
    Exports an OCEL to a SQLite database using Pandas

    Parameters
    ---------------
    ocel
        Object-centric event log
    target_path
        Path to the SQLite database
    parameters
        Parameters of the exporter
    """
    if parameters is None:
        parameters = {}

    import sqlite3

    if os.path.exists(target_path):
        os.remove(target_path)

    ocel = ocel_consistency.apply(ocel, parameters=parameters)
    ocel = filtering_utils.propagate_relations_filtering(
        ocel, parameters=parameters
    )

    conn = sqlite3.connect(target_path)

    ocel.events.to_sql("EVENTS", conn, index=False)
    ocel.relations.to_sql("RELATIONS", conn, index=False)
    ocel.objects.to_sql("OBJECTS", conn, index=False)

    conn.close()



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import logging
from enum import Enum

from pm4py.objects.log.obj import Trace, Event
from pm4py.util import xes_constants, exec_utils
from pm4py.util.dt_parsing import parser as dt_parser


class Parameters(Enum):
    ACCEPTANCE_CONDITION = "acceptance_condition"


_EVENT_END = "end"
_EVENT_START = "start"


def parse_attribute(elem, store, key, value, tree):
    if len(elem.getchildren()) == 0:
        if type(store) is list:
            # changes to the store of lists: not dictionaries anymore
            # but pairs of key-values.
            store.append((key, value))
        else:
            store[key] = value
    else:
        if elem.getchildren()[0].tag.endswith(xes_constants.TAG_VALUES):
            store[key] = {
                xes_constants.KEY_VALUE: value,
                xes_constants.KEY_CHILDREN: list(),
            }
            tree[elem] = store[key][xes_constants.KEY_CHILDREN]
            tree[elem.getchildren()[0]] = tree[elem]
        else:
            store[key] = {
                xes_constants.KEY_VALUE: value,
                xes_constants.KEY_CHILDREN: dict(),
            }
            tree[elem] = store[key][xes_constants.KEY_CHILDREN]
    return tree


class StreamingTraceXesReader:
    def __init__(self, path, parameters=None):
        """
        Initialize the iterable log object

        Parameters
        -------------
        path
            Path to the XES log
        """
        if parameters is None:
            parameters = {}
        self.path = path
        self.acceptance_condition = exec_utils.get_param_value(
            Parameters.ACCEPTANCE_CONDITION, parameters, lambda x: True
        )
        self.date_parser = dt_parser.get()
        self.reset()

    def __iter__(self):
        """
        Starts the iteration
        """
        return self

    def __next__(self):
        """
        Gets the next element of the log
        """
        trace = self.read_trace()
        if self.reading_log:
            return trace
        raise StopIteration

    def to_trace_stream(self, trace_stream):
        """
        Sends the content of a XES log to a trace stream

        Parameters
        --------------
        trace_stream
            Trace stream
        """
        while self.reading_log:
            trace = self.read_trace()
            if trace is not None:
                trace_stream.append(trace)

    def reset(self):
        """
        Resets the iterator
        """
        # reset the variables
        from lxml import etree

        self.context = None
        self.tree = None
        # initialize the variables
        self.context = etree.iterparse(
            self.path, events=[_EVENT_START, _EVENT_END]
        )
        self.trace = None
        self.event = None
        self.reading_log = True
        self.reading_trace = False
        self.tree = {}

    def read_trace(self):
        """
        Gets the next trace from the iterator

        Returns
        ------------
        trace
            Trace
        """
        tree = self.tree
        while True:
            tree_event, elem = next(self.context)

            if tree_event == _EVENT_START:
                parent = (
                    tree[elem.getparent()]
                    if elem.getparent() in tree
                    else None
                )

                if elem.tag.endswith(xes_constants.TAG_TRACE):
                    self.trace = Trace()
                    tree[elem] = self.trace.attributes
                    self.reading_trace = True
                    continue

                if self.reading_trace:
                    if elem.tag.endswith(xes_constants.TAG_STRING):
                        if parent is not None:
                            tree = parse_attribute(
                                elem,
                                parent,
                                elem.get(xes_constants.KEY_KEY),
                                elem.get(xes_constants.KEY_VALUE),
                                tree,
                            )
                        continue

                    elif elem.tag.endswith(xes_constants.TAG_DATE):
                        try:
                            dt = self.date_parser.apply(
                                elem.get(xes_constants.KEY_VALUE)
                            )
                            tree = parse_attribute(
                                elem,
                                parent,
                                elem.get(xes_constants.KEY_KEY),
                                dt,
                                tree,
                            )
                        except TypeError:
                            logging.info(
                                "failed to parse date: "
                                + str(elem.get(xes_constants.KEY_VALUE))
                            )
                        except ValueError:
                            logging.info(
                                "failed to parse date: "
                                + str(elem.get(xes_constants.KEY_VALUE))
                            )
                        continue

                    elif elem.tag.endswith(xes_constants.TAG_EVENT):
                        self.event = Event()
                        tree[elem] = self.event
                        continue

                    elif elem.tag.endswith(xes_constants.TAG_FLOAT):
                        if parent is not None:
                            try:
                                val = float(elem.get(xes_constants.KEY_VALUE))
                                tree = parse_attribute(
                                    elem,
                                    parent,
                                    elem.get(xes_constants.KEY_KEY),
                                    val,
                                    tree,
                                )
                            except ValueError:
                                logging.info(
                                    "failed to parse float: "
                                    + str(elem.get(xes_constants.KEY_VALUE))
                                )
                        continue

                    elif elem.tag.endswith(xes_constants.TAG_INT):
                        if parent is not None:
                            try:
                                val = int(elem.get(xes_constants.KEY_VALUE))
                                tree = parse_attribute(
                                    elem,
                                    parent,
                                    elem.get(xes_constants.KEY_KEY),
                                    val,
                                    tree,
                                )
                            except ValueError:
                                logging.info(
                                    "failed to parse int: "
                                    + str(elem.get(xes_constants.KEY_VALUE))
                                )
                        continue

                    elif elem.tag.endswith(xes_constants.TAG_BOOLEAN):
                        if parent is not None:
                            try:
                                val0 = elem.get(xes_constants.KEY_VALUE)
                                val = False
                                if str(val0).lower() == "true":
                                    val = True
                                tree = parse_attribute(
                                    elem,
                                    parent,
                                    elem.get(xes_constants.KEY_KEY),
                                    val,
                                    tree,
                                )
                            except ValueError:
                                logging.info(
                                    "failed to parse boolean: "
                                    + str(elem.get(xes_constants.KEY_VALUE))
                                )
                        continue

                    elif elem.tag.endswith(xes_constants.TAG_LIST):
                        if parent is not None:
                            # lists have no value, hence we put None as a value
                            tree = parse_attribute(
                                elem,
                                parent,
                                elem.get(xes_constants.KEY_KEY),
                                None,
                                tree,
                            )
                        continue

                    elif elem.tag.endswith(xes_constants.TAG_ID):
                        if parent is not None:
                            tree = parse_attribute(
                                elem,
                                parent,
                                elem.get(xes_constants.KEY_KEY),
                                elem.get(xes_constants.KEY_VALUE),
                                tree,
                            )
                        continue

            elif tree_event == _EVENT_END:
                if elem in tree:
                    del tree[elem]
                elem.clear()
                if elem.getprevious() is not None:
                    try:
                        del elem.getparent()[0]
                    except TypeError:
                        pass

                if elem.tag.endswith(xes_constants.TAG_EVENT):
                    if self.trace is not None:
                        self.trace.append(self.event)
                        self.event = None
                    continue

                elif elem.tag.endswith(xes_constants.TAG_TRACE):
                    self.reading_trace = False
                    if self.acceptance_condition(self.trace):
                        return self.trace
                    continue

                elif elem.tag.endswith(xes_constants.TAG_LOG):
                    self.reading_log = False
                    break


def apply(path, parameters=None):
    """
    Creates a StreamingTraceXesReader object

    Parameters
    ---------------
    path
        Path
    parameters
        Parameters of the algorithm

    Returns
    ---------------
    stream_read_obj
        Stream reader object
    """
    return StreamingTraceXesReader(path, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Dict, Any
from enum import Enum
from typing import Optional
from pm4py.objects.ocel.importer.sqlite.variants import pandas_importer, ocel20
from pm4py.util import exec_utils


class Variants(Enum):
    PANDAS_IMPORTER = pandas_importer
    OCEL20 = ocel20


def apply(
    file_path: str,
    variant=Variants.PANDAS_IMPORTER,
    parameters: Optional[Dict[Any, Any]] = None,
) -> OCEL:
    """
    Imports an OCEL from a SQLite database

    Parameters
    --------------
    file_path
        Path to the SQLite database
    variant
        Variant of the importer to use:
        - Variants.PANDAS_IMPORTER => Pandas
    parameters
        Variant-specific parameters

    Returns
    --------------
    ocel
        Object-centric event log
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(
        file_path, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from abc import ABC
from collections import Counter
from typing import List, Collection, Any, Optional, Generic, Dict

from pm4py.algo.discovery.inductive.cuts.abc import Cut, T
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureUVCL,
    IMDataStructureDFG,
)
from pm4py.objects.dfg import util as dfu
from pm4py.objects.dfg.obj import DFG
from pm4py.objects.process_tree.obj import Operator, ProcessTree


class ConcurrencyCut(Cut[T], ABC, Generic[T]):

    @classmethod
    def operator(
        cls, parameters: Optional[Dict[str, Any]] = None
    ) -> ProcessTree:
        return ProcessTree(operator=Operator.PARALLEL)

    @classmethod
    def holds(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[List[Collection[Any]]]:
        dfg = obj.dfg
        alphabet = dfu.get_vertices(dfg)
        alphabet = sorted(list(alphabet))
        edges = dfu.get_edges(dfg)
        edges = sorted(list(edges))

        groups = [{a} for a in alphabet]
        if len(groups) == 0:
            return None

        cont = True
        while cont:
            cont = False
            i = 0
            while i < len(groups):
                j = i + 1
                while j < len(groups):
                    for act1 in groups[i]:
                        for act2 in groups[j]:
                            if (act1, act2) not in edges or (
                                act2,
                                act1,
                            ) not in edges:
                                groups[i] = groups[i].union(groups[j])
                                del groups[j]
                                cont = True
                                break
                        if cont:
                            break
                    if cont:
                        break
                    j = j + 1
                if cont:
                    break
                i = i + 1

        groups = list(sorted(groups, key=lambda g: len(g)))
        i = 0
        while i < len(groups) and len(groups) > 1:
            if (
                len(groups[i].intersection(set(dfg.start_activities.keys())))
                > 0
                and len(groups[i].intersection(set(dfg.end_activities.keys())))
                > 0
            ):
                i += 1
                continue
            group = groups[i]
            del groups[i]
            if i == 0:
                groups[i].update(group)
            else:
                groups[i - 1].update(group)

        return groups if len(groups) > 1 else None


class ConcurrencyCutUVCL(ConcurrencyCut[IMDataStructureUVCL]):

    @classmethod
    def project(
        cls,
        obj: IMDataStructureUVCL,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureUVCL]:
        r = list()
        for g in groups:
            c = Counter()
            for t in obj.data_structure:
                c[tuple(filter(lambda e: e in g, t))] = obj.data_structure[t]
            r.append(c)
        return list(map(lambda l: IMDataStructureUVCL(l), r))


class ConcurrencyCutDFG(ConcurrencyCut[IMDataStructureDFG]):

    @classmethod
    def project(
        cls,
        obj: IMDataStructureDFG,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureDFG]:
        dfgs = []
        skippable = []
        for g in groups:
            dfn = DFG()
            for a in obj.dfg.start_activities:
                if a in g:
                    dfn.start_activities[a] = obj.dfg.start_activities[a]
            for a in obj.dfg.end_activities:
                if a in g:
                    dfn.end_activities[a] = obj.dfg.end_activities[a]
            for a, b in obj.dfg.graph:
                if a in g and b in g:
                    dfn.graph[(a, b)] = obj.dfg.graph[(a, b)]
            skippable.append(False)
            dfgs.append(dfn)
        r = list()
        [
            r.append(
                IMDataStructureDFG(
                    InductiveDFG(dfg=dfgs[i], skip=skippable[i])
                )
            )
            for i in range(len(dfgs))
        ]
        return r
